using System;
using System.Text;
using System.Runtime.InteropServices;
using System.Collections;

namespace AgileSoftwareInc.Tapi2Lib
{


	public sealed class LocationCollection : IEnumerable
	{
		/*
			class LocationCollection
			Copyright  2003 by Best Software - CRM Divison.
			All Rights Reserved.
      
			Created: Brian Hormann
			Updated: February 22, 2003

		*/
		private ArrayList m_locations;

		internal LocationCollection()
		{
			m_locations = new ArrayList();
		}
		internal LocationCollection(int numberLocations)
		{
			m_locations = new ArrayList(numberLocations);
		}

		public IEnumerator GetEnumerator()
		{
			return m_locations.GetEnumerator();
		}

		public void Add(Location location)
		{
			m_locations.Add(location);
		}

		public int Count 
		{
			get { return m_locations.Count; }
		}

		public void Clear()
		{
			m_locations.Clear();
		}

		public void Remove(Location location)
		{
			m_locations.Remove(location);
		}

		public bool Contains(Location location)
		{
			return m_locations.Contains(location);
		}
		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public Location[] ToArray()
		{	
			return (Location[])this.m_locations.ToArray(typeof(Location));
		}
	} // class LocationCollection

	/// <summary>
	/// Summary description for Location.
	/// </summary>
	public class Location
	{
		/*
			class Location
			Copyright  2003 by Best Software - CRM Divison.
			All Rights Reserved.
      
			Created: Brian Hormann
			Updated: February 22, 2003

		*/
		#region Private Members

		internal int		permanentID;
		internal string name;
		internal int		countryCode;
		internal string cityCode;
		internal int		preferredCardID;
		internal string localAccessCode;
		internal string longDistanceAccessCode;
		internal string tollPrefixList;
		internal int		countryID;
		internal int		options;
		internal string cancelCallWaiting;

		#endregion

		#region Constr's
		public Location()
		{}

		#endregion

		#region Public
		/// <summary>
		/// Permanent identifier that identifies the location.
		/// </summary>
		public int PermanentID
		{
			get{return permanentID;}
		}
		/// <summary>
		///	string that describes the location in a user-friendly manner
		/// </summary>
		public string Name
		{
			get{return name;}
		}
		/// <summary>
		/// Country code of the location
		/// </summary>
		public int CountryCode
		{
			get{return countryCode;}
		}
		/// <summary>
		/// string specifying the city/area code associated with the location. 
		/// This information, along with the country code, can be used by 
		/// applications to "default" entry fields for the user when 
		/// entering phone numbers, to encourage the entry of proper 
		/// canonical numbers.
		/// </summary>
		public string CityCode
		{
			get{return cityCode;}
		}
		/// <summary>
		/// Preferred calling card when dialing from this location. 
		/// </summary>
		public int PreferredCardID
		{
			get{return preferredCardID;}
		}
		/// <summary>
		/// string containing the access code to be 
		/// dialed before calls to addresses in the local calling area.
		/// </summary>
		public string LocalAccessCode
		{
			get{return localAccessCode;}
		}
		/// <summary>
		/// string containing the access code to be dialed before 
		/// calls to addresses outside the local calling area. 
		/// </summary>
		public string LongDistanceAccessCode
		{
			get{return longDistanceAccessCode;}
		}

		/// <summary>
		/// string containing the toll prefix list for the location. 
		/// The string contains only prefixes consisting of the 
		/// digits "0" through "9", separated from each other 
		/// by a single "," (comma) character.
		/// </summary>
		public string TollPrefixList
		{
			get{return tollPrefixList;}
		}
		/// <summary>
		/// Country identifier of the country/region selected 
		/// for the location. This can be used with the lineGetCountry 
		/// function to obtain additional information about the 
		/// specific country/region, such as the country/region name 
		/// </summary>
		public int CountryID
		{
			get{return countryID;}
		}
		/// <summary>
		/// Options in effect for this location
		/// </summary>
		public int Options
		{
			get{return options;}
		}
		/// <summary>
		/// string containing the dial digits and modifier characters 
		/// that should be prefixed to the dialable string (after 
		/// the pulse/tone character) when an application sets the 
		/// LINETRANSLATEOPTION_CANCELCALLWAITING bit in the 
		/// dwTranslateOptions parameter of lineTranslateAddress
		/// </summary>
		public string CancelCallWaiting
		{
			get{return cancelCallWaiting;}
		}

		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			return name;
		}

		public string GetInfoString()
		{
			StringBuilder sb = new StringBuilder();
			sb.Append("PermanentID = " + permanentID.ToString() + Environment.NewLine);
			sb.Append("Name = " + name.ToString() + Environment.NewLine);
			sb.Append("Country Code = " + countryCode + Environment.NewLine);
			sb.Append("City Code = " + cityCode + Environment.NewLine);
			sb.Append("Preferred Card ID = " + preferredCardID + Environment.NewLine);
			sb.Append("Local Access Code = " + localAccessCode + Environment.NewLine);
			sb.Append("Long Distance Access Code = " + preferredCardID + Environment.NewLine);
			sb.Append("Toll Prefix List = " + tollPrefixList + Environment.NewLine);
			sb.Append("Country ID = " + countryID + Environment.NewLine);
			sb.Append("Options = " + options + Environment.NewLine);
			sb.Append("Cancel Call Waiting = " + cancelCallWaiting + Environment.NewLine);
			return sb.ToString();
		}

		#endregion
	} // class Location
}// namespace AgileSoftwareInc.Tapi2Lib



