#define TAPI22
//#define TESTING

using System;
using System.Runtime.InteropServices;
using System.Text;
// for debugging -- showing MessageBoxes
using System.Windows.Forms;
// for debugging
using System.Diagnostics;

namespace AgileSoftwareInc.Tapi2Lib
/*
  namespace Tapi2Lib contains all classes for C# wrapper for Tapi 2
  Copyright  2003 -- 2008 by Agile Software Inc.
  All Rights Reserved.
  
  Created: Helen Warn
 * Updated: jun 16, 2008 to add lineGenerateDigits + events
 * Updated: Feb 12, 2008 to add lineSendUserUserInfo
 * Updated: Feb 9, 2008 to use Net 2 classes
 * Updated: June 20, 2006 to add fix from gotdotnet to MakeCall()
 * Updated: June 19, 2006 to add UnPark
 * Updated: May 4, 2006 to add Hold, Unhold
	Updated: January 24, 2006 to add LINE_CALLINFO event handling
	Updated	: April 1, 2003 by Helen Warn
  Updated: February 22, 2003
	Updated: Christoph Brndle
	Modified: Paul Basnett
	Mods	: Now makes and answer calls via modems.
*/
{

	/// <summary>
	/// CTapi is where we keep all our structs, etc. It is the parent class for all Tapi objects
	/// and contains the initialization of Tapi & a function for choosing a line to open 
	/// </summary>
	public class CTapi: IDisposable
	{
    /*
      class CTapi
      Copyright  2003 - 2008 by Agile Software Inc.
      All Rights Reserved.
      
      Created: Helen Warn
     * Updated: 2008 to add more api prototypes
			Updated: January 24, 2006 to add LINE_CALLINFO event handling
			Updated: Christoph Brandle & Brian Hormann
      Updated: February 22, 2003

    */		
		#region Constants
		
		public const uint TAPILOWVERSION = (uint)0x10003;
		public const uint TAPIHIVERSION = (uint) 0x20002;
		public const uint TSPLOWEXTVERSION = (uint) 0x10001;
		public const uint TSPHIEXTVERSION = (uint) 0x10001;
		
		#endregion // Constants

		#region EventArgs
	  
		/// <summary>
		/// for event LINE_APPNEWCALL
		/// </summary>
		public class AppNewCallEventArgs: EventArgs
		{
			public readonly CLine line;
			public readonly CCall call;
	    
			public AppNewCallEventArgs (CLine Line, CCall Call)
			{
				line = Line;
				call = Call;
			}
	    
		}
  
		/// <summary>
		/// for event LINE_CALLSTATE
		/// </summary>
		public class CallStateEventArgs: EventArgs
		{
			// we need the arg hcall to handle the IDLE message. In this call the call is null but
			// the call handle is passed.
			public readonly IntPtr hcall;
			public readonly CCall call;
			public readonly LineCallState CallState;
			public readonly uint CallStateDetail;
	    
			public CallStateEventArgs (IntPtr hCall, CCall Call, LineCallState callState, uint callStateDetail)
			{
				hcall = hCall;
				call = Call;
				CallState = callState;
				CallStateDetail = callStateDetail;
			}
		}
  
		/// <summary>
		/// for event LINE_REPLY
		/// </summary>
		public class LineReplyEventArgs: EventArgs
		{
			public readonly uint IDRequest;
			public readonly uint status;
	    
			public LineReplyEventArgs(uint request, uint stat)
			{
				IDRequest = request;
				status = stat;
			}
	  
		}

		/// <summary>
		/// for event LINE_CALLINFO
		/// </summary>
		public class LineCallInfoEventArgs: EventArgs
		{
			public readonly IntPtr hCall;
			public readonly LineCallInfoState callInfoState;

			public LineCallInfoEventArgs(IntPtr hcall, LineCallInfoState callinfo)
			{
				hCall = hcall;
				callInfoState = callinfo;
			}
		}

    /// <summary>
    /// for event LINE_GATHERTERM
    /// </summary>
    public class LineGatherDigitsEventArgs : EventArgs
    {
      public readonly IntPtr hCall;
      public readonly LineGatherTerm lineGatherTerm;
      public LineGatherDigitsEventArgs(IntPtr hcall, LineGatherTerm gatherTerm)
			{
				hCall = hcall;
        lineGatherTerm = gatherTerm;
			}
    }
  
		#endregion // EventArgs
  
		#region Events
  
		public event AppNewCallEventHandler AppNewCall = null;
		public event CallStateEventHandler CallStateEvent = null;
		public event LineReplyEventHandler LineReplyEvent = null;
		public event LineCallInfoEventHandler LineCallInfoEvent = null;
    public event LineGatherDigitsEventHandler LineGatherDigits = null;
  
		#endregion // Events
  
		#region Delegates
  
		/// <summary>
		/// main callback which is the entry to all Tapi Events
		/// </summary>
		public delegate void LineCallBack(System.UInt32 dwDevice, System.UInt32 dwMessage, System.UInt32 dwInstance,
			System.UInt32 dwParam1, System.UInt32 dwParam2, System.UInt32 dwParam3);
	
		/// <summary>
		/// new call
		/// </summary>
		public delegate void AppNewCallEventHandler(Object sender, AppNewCallEventArgs e);
	
		/// <summary>
		/// LINE_CALLSTATE event handler
		/// </summary>
		public delegate void CallStateEventHandler(Object sender, CallStateEventArgs e);
	
		/// <summary>
		/// LINE_REPLY event handler
		/// </summary>
		public delegate void LineReplyEventHandler(Object sender, LineReplyEventArgs e);


		/// <summary>
		/// LINE_CALLINFO event handler
		/// </summary>
		public delegate void LineCallInfoEventHandler(Object sender, LineCallInfoEventArgs e);

    /// <summary>
    /// LINE_GATHERDIGITS event handler
    /// </summary>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    public delegate void LineGatherDigitsEventHandler(Object sender, LineGatherDigitsEventArgs e);
	
		#endregion //Delegates
	  		
		#region API Calls		
		// make sure case is exact for all dll imports!!!

		private const string TAPILIB = "tapi32.dll";

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern CTapi.LineErrReturn lineAccept(
			IntPtr hCall,
			byte[] UserUserInfo,
			uint dwSize);

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern CTapi.LineErrReturn lineAnswer(
			IntPtr hCall,
			byte[] UserUserInfo,
			uint dwSize);

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern int lineBlindTransfer(
			IntPtr hCall,
			string lpszDestAddress,
			uint dwCountryCode);
	
		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]		
		internal static extern CTapi.LineErrReturn lineClose(
			IntPtr hLine);


		[DllImport(TAPILIB, EntryPoint="lineConfigDialogW", CharSet=CharSet.Auto)]
		internal static extern LineErrReturn lineConfigDialog(uint dwDeviceID, 
			System.IntPtr hwndOwner, string lpszDeviceClass);

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]		
		internal static extern CTapi.LineErrReturn lineDeallocateCall(
			IntPtr hLine);

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]	
		internal static extern int lineDevSpecific(
			IntPtr hLine,               
			uint dwAddressID, 
			IntPtr hCall,              
			IntPtr lpParams,              
			uint dwSize);	

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]	
		internal static extern int lineDrop(
			IntPtr	hCall,
			string	UserInfo,
			uint	dwSize);

    [DllImport(TAPILIB, CharSet = CharSet.Auto, SetLastError = true)]
    internal static extern CTapi.LineErrReturn lineGenerateDigits(
      IntPtr hCall,
      LineDigitMode dwDigitMode,
      string lpszDigits,
      uint dwDuration);
  
		[DllImport(TAPILIB, EntryPoint="lineGetAddressCapsW", CharSet=CharSet.Auto, SetLastError=true)]	
		internal static extern CTapi.LineErrReturn lineGetAddressCaps(
			IntPtr hLineApp,               
			uint dwDeviceID,                
			uint dwAddressID,               
			uint dwAPIVersion,              
			uint dwExtVersion,              
			IntPtr lpAddressCaps);	

		[DllImport(TAPILIB, EntryPoint="lineGetCallInfoW", CharSet=CharSet.Auto, SetLastError=true)]	
		internal static extern CTapi.LineErrReturn lineGetCallInfo(
			IntPtr hCall,                       
			IntPtr lpCallInfo);	

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern CTapi.LineErrReturn lineGetCallStatus(IntPtr hCall, IntPtr lpCallStatus);

		/// <summary>
		/// The lineGetCountry function fetches the stored dialing rules 
		/// and other information related to a specified country/region, 
		/// the first country/region in the country/region list, 
		/// or all countries/regions.
		/// </summary>
		/// <param name="dwCountryID">Country/region identifier 
		/// (not the country code) of the country/region for which 
		/// information is to be obtained. If the value 1 is specified, 
		/// information on the first country/region in the 
		/// country/region list is obtained. If the value 0 is specified, 
		/// information on all countries/regions is obtained 
		/// (which can require a great deal of memory  20 KB or more).</param>
		/// <param name="dwAPIVersion">Highest version of TAPI supported by 
		/// the application (not necessarily the value negotiated 
		/// by lineNegotiateAPIVersion on some particular line device). </param>
		/// <param name="lpLineCountryList">Pointer to a location to 
		/// which a LINECOUNTRYLIST structure is loaded. Prior to 
		/// calling lineGetCountry, the application must set the 
		/// dwTotalSize member of this structure to indicate the amount 
		/// of memory available to TAPI for returning information</param>
		/// <returns>Returns LineErrReturn.OK if the request succeeds or a negative error 
		/// number if an error occurs. Possible return values are:		
		/// LineErrReturn.INCOMPATIBLEAPIVERSION, LineErrReturn.NOMEM, 
		/// LineErrReturn.INIFILECORRUPT, LineErrReturn.OPERATIONFAILED, 
		/// LineErrReturn.INVALCOUNTRYCODE, LineErrReturn.STRUCTURETOOSMALL, 
		/// LineErrReturn.INVALPOINTER. <see cref="LineErrReturn"/></returns>
		[DllImport(TAPILIB, EntryPoint="lineGetCountryW", CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern CTapi.LineErrReturn lineGetCountry(int dwCountryID, int dwAPIVersion, 
			IntPtr lpLineCountryList);
 
		[DllImport(TAPILIB, EntryPoint="lineGetDevCapsW", CharSet=CharSet.Auto, SetLastError=true)]		
		internal static extern CTapi.LineErrReturn lineGetDevCaps(
			IntPtr hLineApp,           
			uint dwDeviceID,            
			uint dwAPIVersion,          
			uint dwExtVersion,          
			IntPtr lpLineDevCaps);

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern CTapi.LineErrReturn lineGetDevConfig(
			uint dwDeviceID,
			ref varstring DeviceConfig,
			string DeviceClass);

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern CTapi.LineErrReturn lineGetID(
			IntPtr hLine,
			uint dwAddressID,
			IntPtr hCall,
			uint dwSelect,
			out varstring DeviceID,
			string DeviceClass);

		[DllImport(TAPILIB, EntryPoint="lineGetTranslateCapsW", CharSet=CharSet.Auto, SetLastError=true)]	
		internal static extern CTapi.LineErrReturn 
			lineGetTranslateCaps(IntPtr hLineApp, uint dwAPIVersion, IntPtr lpTranslateCaps);

		[DllImport(TAPILIB, EntryPoint = "lineHold", CharSet = CharSet.Auto, SetLastError = true)]
		internal static extern int LineHold(IntPtr hCall);

		[DllImport(TAPILIB, EntryPoint="lineInitializeW", CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern CTapi.LineErrReturn lineInitialize(
			out IntPtr hLineApp, 
			IntPtr hAppHandle, 
			LineCallBack lCalllBack,
			string FriendlyAppName, 
			out System.UInt32 NumDevices, 
			ref System.UInt32 APIVersion);
      		  	
		[DllImport(TAPILIB, EntryPoint="lineInitializeExW", CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern CTapi.LineErrReturn lineInitializeEx(
			out IntPtr hLineApp, 
			IntPtr hAppHandle, 
			LineCallBack lCalllBack,
			string FriendlyAppName, 
			out System.UInt32 NumDevices, 
			ref System.UInt32 APIVersion,
			ref CTapi.lineinitializeexparams lineExInitParams);

		[DllImport(TAPILIB, EntryPoint="lineMakeCallW", CharSet=CharSet.Auto, SetLastError=true)]	
		internal static extern int lineMakeCall(
			IntPtr hLine,               
			out IntPtr hCall,
			string DestAddress,
			uint CountryCode,
			IntPtr lpCallParams);

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]		
		internal static extern CTapi.LineErrReturn lineNegotiateAPIVersion(
			IntPtr hLineApp,               
			uint dwDeviceID,                
			uint dwAPILowVersion,           
			uint dwAPIHighVersion,          
			out uint lpdwAPIVersion,          
			out CTapi.lineextensionid lpExtensionID); 
   
		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]		
		internal static extern CTapi.LineErrReturn lineNegotiateExtVersion(
			IntPtr hLineApp,               
			uint dwDeviceID,                
			uint dwAPIVersion,          
			uint dwExtLowVersion,           
			uint dwExtHighVersion,          
			out  uint dwExtVersion); 

		[DllImport(TAPILIB, EntryPoint="lineOpenW", CharSet=CharSet.Auto, SetLastError=true)]		
		internal static extern CTapi.LineErrReturn lineOpen(
			IntPtr hLineApp, 
			uint dwDeviceID,
			out IntPtr hLine,
			uint dwAPIVersion,
			uint dwExtVersion,
			LineCallBack dwCallbackInstance,
			LineCallPrivilege dwPrivileges,
			LineMediaMode dwMediaModes,
			linecallparams lpCallParams);

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]		 
		internal static extern CTapi.LineErrReturn linePark(
			IntPtr hCall,           
			LineParkMode dwParkMode,
			string lpszDirAddress,
			out varstring lpNonDirAddress
			);

    [DllImport(TAPILIB, CharSet = CharSet.Auto, SetLastError = true)]
    internal static extern CTapi.LineErrReturn lineReleaseUserUserInfo(
      IntPtr hCall
      );

    [DllImport(TAPILIB, CharSet = CharSet.Auto, SetLastError = true)]
    internal static extern int lineSendUserUserInfo(
      IntPtr hCall,
      String lpszUserUserInfo,
      uint dwSize
      );

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]		 
		internal static extern CTapi.LineErrReturn lineSetCallPrivilege(
			IntPtr hCall,           
			LineCallPrivilege dwCallPrivilege  
			);

		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]		 
		internal static extern CTapi.LineErrReturn lineSetStatusMessages(
			IntPtr hLine,           
			LineDevState dwLineStates,    
			LineAddressState dwAddressStates  
			);
	  		  
		[DllImport(TAPILIB, CharSet=CharSet.Auto, SetLastError=true)]
		internal static extern CTapi.LineErrReturn lineShutdown(System.IntPtr hTapi);		

		[DllImport(TAPILIB, EntryPoint="lineTranslateAddressW", CharSet=CharSet.Auto, SetLastError=true)]	
		internal static extern CTapi.LineErrReturn lineTranslateAddress(
			IntPtr hLineApp,
			uint dwDeviceID,
			uint dwAPIVersion,
			string AddressIn,
			uint dwCard,
			uint dwTranslateOptions,
			IntPtr lpTranslateOutput);


		[DllImport(TAPILIB, EntryPoint="lineTranslateDialogW", CharSet=CharSet.Auto)]
		internal static extern LineErrReturn lineTranslateDialog(System.IntPtr hLineApp, 
			uint dwDeviceID, uint dwAPIVersion, IntPtr hwndOwner, string lpszAddressIn);

		[DllImport(TAPILIB, EntryPoint = "lineUnhold", CharSet = CharSet.Auto, SetLastError = true)]
		internal static extern int LineUnhold(IntPtr hCall);

		[DllImport(TAPILIB, CharSet = CharSet.Auto, SetLastError = true)]
		internal static extern CTapi.LineErrReturn lineUnpark(
			IntPtr hLine,
			uint dwAddressID,
			out IntPtr hCall,
			string lpszDirAddress
			);

		#endregion // API Calls
	
		#region structs & helper classes

		[StructLayout(LayoutKind.Sequential)]     
			public class lineaddresscaps 
		{
			public uint  dwTotalSize;
			public uint  dwNeededSize;
			public uint  dwUsedSize;

			public uint  dwLineDeviceID;

			public uint  dwAddressSize;
			public uint  dwAddressOffset;

			public uint  dwDevSpecificSize;
			public uint  dwDevSpecificOffset;

			public LineAddressSharing  dwAddressSharing;
			public LineAddressState dwAddressStates;
			public LineCallInfoState  dwCallInfoStates;
			public LineCallPartyID  dwCallerIDFlags;
			public LineCallPartyID  dwCalledIDFlags;
			public LineCallPartyID  dwConnectedIDFlags;
			public LineCallPartyID  dwRedirectionIDFlags;
			public LineCallPartyID  dwRedirectingIDFlags;
			public LineCallState  dwCallStates;
			public LineDialToneMode  dwDialToneModes;
			public LineBusyMode  dwBusyModes;
			public LineSpecialInfo  dwSpecialInfo;
			public LineDisconnectedMode  dwDisconnectModes;

			public uint  dwMaxNumActiveCalls;
			public uint  dwMaxNumOnHoldCalls;
			public uint  dwMaxNumOnHoldPendingCalls;
			public uint  dwMaxNumConference;
			public uint  dwMaxNumTransConf;

			public LineAddrCapFlags  dwAddrCapFlags;
			public LineCallFeature  dwCallFeatures;
			public LineRemoveFromConf  dwRemoveFromConfCaps;
			public LineCallState  dwRemoveFromConfState;
			public LineTransferMode  dwTransferModes;
			public LineParkMode dwParkModes;

			public LineForwardMode  dwForwardModes;
			public uint  dwMaxForwardEntries;
			public uint  dwMaxSpecificEntries;
			public uint  dwMinFwdNumRings;
			public uint  dwMaxFwdNumRings;

			public uint  dwMaxCallCompletions;
			public uint  dwCallCompletionConds;
			public uint  dwCallCompletionModes;
			public uint  dwNumCompletionMessages;
			public uint  dwCompletionMsgTextEntrySize;
			public uint  dwCompletionMsgTextSize;
			public uint  dwCompletionMsgTextOffset;
			public LineAddrFeature  dwAddressFeatures;

			public uint  dwPredictiveAutoTransferStates;
			public uint  dwNumCallTreatments;
			public uint  dwCallTreatmentListSize;
			public uint  dwCallTreatmentListOffset;
			public uint  dwDeviceClassesSize;
			public uint  dwDeviceClassesOffset;
			public uint  dwMaxCallDataSize;
			public LineCallFeature2  dwCallFeatures2;
			public uint  dwMaxNoAnswerTimeout;
			public LineConnectedMode  dwConnectedModes;
			public LineOfferingMode  dwOfferingModes;
			public LineMediaMode  dwAvailableMediaModes;     
			/*
			[MarshalAs(UnmanagedType.ByValArray, SizeConst=1 )] 
			public byte[] buffer; // set to proper size to handle variable-length
			*/
		}

   
		[StructLayout(LayoutKind.Sequential)]     
			public class linecallinfo 
		{
			public uint  dwTotalSize;
			public uint  dwNeededSize;
			public uint  dwUsedSize;

			public System.IntPtr hLine;
			public uint  dwLineDeviceID;
			public uint  dwAddressID;

			public uint  dwBearerMode;
			public uint  dwRate;
			public uint  dwMediaMode;

			public uint  dwAppSpecific;
			public uint  dwCallID;
			public uint  dwRelatedCallID;
			public uint  dwCallParamFlags;
			public uint  dwCallStates;

			public uint  dwMonitorDigitModes;
			public uint  dwMonitorMediaModes;
    
			public linedialparams  DialParams;

			public uint  dwOrigin;
			public uint  dwReason;
			public uint  dwCompletionID;

			public uint  dwNumOwners;
			public uint  dwNumMonitors;

			public uint  dwCountryCode;
			public uint  dwTrunk;

			public uint  dwCallerIDFlags;
			public uint  dwCallerIDSize;
			public uint  dwCallerIDOffset;
			public uint  dwCallerIDNameSize;
			public uint  dwCallerIDNameOffset;

			public uint  dwCalledIDFlags;
			public uint  dwCalledIDSize;
			public uint  dwCalledIDOffset;
			public uint  dwCalledIDNameSize;
			public uint  dwCalledIDNameOffset;

			public uint  dwConnectedIDFlags;
			public uint  dwConnectedIDSize;
			public uint  dwConnectedIDOffset;
			public uint  dwConnectedIDNameSize;
			public uint  dwConnectedIDNameOffset;

			public uint  dwRedirectionIDFlags;
			public uint  dwRedirectionIDSize;
			public uint  dwRedirectionIDOffset;
			public uint  dwRedirectionIDNameSize;
			public uint  dwRedirectionIDNameOffset;

			public uint  dwRedirectingIDFlags;
			public uint  dwRedirectingIDSize;
			public uint  dwRedirectingIDOffset;
			public uint  dwRedirectingIDNameSize;
			public uint  dwRedirectingIDNameOffset;

			public uint  dwAppNameSize;
			public uint  dwAppNameOffset;
			public uint  dwDisplayableAddressSize; 
			public uint  dwDisplayableAddressOffset; 
   
			public uint  dwCalledPartySize;
			public uint  dwCalledPartyOffset;

			public uint  dwCommentSize;
			public uint  dwCommentOffset;

			public uint  dwDisplaySize;
			public uint  dwDisplayOffset;

			public uint  dwUserUserInfoSize;
			public uint  dwUserUserInfoOffset;

			public uint  dwHighLevelCompSize;
			public uint  dwHighLevelCompOffset;

			public uint  dwLowLevelCompSize;
			public uint  dwLowLevelCompOffset;

			public uint  dwChargingInfoSize;
			public uint  dwChargingInfoOffset;

			public uint  dwTerminalModesSize;
			public uint  dwTerminalModesOffset;

			public uint  dwDevSpecificSize;
			public uint  dwDevSpecificOffset;

			public uint  dwCallTreatment;

			public uint  dwCallDataSize;
			public uint  dwCallDataOffset;

			public uint  dwSendingFlowspecSize;
			public uint  dwSendingFlowspecOffset;

			public uint  dwReceivingFlowspecSize;
			public uint  dwReceivingFlowspecOffset;

			/*
			[MarshalAs(UnmanagedType.ByValArray, SizeConst=1 )] 
			public byte[] buffer; // set to proper size to handle variable-length struct
			*/
		} 


		[StructLayout(LayoutKind.Sequential)]         
			public class linecallparams
		{  // Defaults:
			public uint  dwTotalSize;          
    
			public uint  dwBearerMode;              // voice
			public uint  dwMinRate;                 // (3.1kHz)
			public uint  dwMaxRate;                 // (3.1kHz)
			public uint  dwMediaMode;               // interactiveVoice

			public uint  dwCallParamFlags;          // 0
			public uint  dwAddressMode;             // addressID
			public uint  dwAddressID;               // (any available)

			public linedialparams DialParams;        // (0, 0, 0, 0)

			public uint  dwOrigAddressSize;         // 0
			public uint  dwOrigAddressOffset;

			public uint  dwDisplayableAddressSize;  // 0
			public uint  dwDisplayableAddressOffset;

			public uint  dwCalledPartySize;         // 0
			public uint  dwCalledPartyOffset;

			public uint  dwCommentSize;             // 0
			public uint  dwCommentOffset;

			public uint  dwUserUserInfoSize;        // 0
			public uint  dwUserUserInfoOffset;

			public uint  dwHighLevelCompSize;       // 0
			public uint  dwHighLevelCompOffset;

			public uint  dwLowLevelCompSize;        // 0
			public uint  dwLowLevelCompOffset;

			public uint  dwDevSpecificSize;         // 0
			public uint  dwDevSpecificOffset;

			public uint  dwPredictiveAutoTransferStates;//TAPI Version 2.0

			public uint  dwTargetAddressSize;    //TAPI Version 2.0
			public uint  dwTargetAddressOffset;  //TAPI Version 2.0

			public uint  dwSendingFlowspecSize;   //TAPI Version 2.0
			public uint  dwSendingFlowspecOffset; //TAPI Version 2.0

			public uint  dwReceivingFlowspecSize;   //TAPI Version 2.0
			public uint  dwReceivingFlowspecOffset; //TAPI Version 2.0

			public uint  dwDeviceClassSize;      //TAPI Version 2.0
			public uint  dwDeviceClassOffset;    //TAPI Version 2.0

			public uint  dwDeviceConfigSize;     //TAPI Version 2.0
			public uint  dwDeviceConfigOffset;   //TAPI Version 2.0

			public uint  dwCallDataSize;         //TAPI Version 2.0
			public uint  dwCallDataOffset;       //TAPI Version 2.0

			public uint  dwNoAnswerTimeout;      //TAPI Version 2.0

			public uint  dwCallingPartyIDSize;   //TAPI Version 2.0
			public uint  dwCallingPartyIDOffset; //TAPI Version 2.0
			/*
			[MarshalAs(UnmanagedType.ByValArray, SizeConst=1 )] 
			public byte[] buffer; // size will be set by lineopen to accommodate VLS
			*/
		}    

		[StructLayout(LayoutKind.Sequential)]		
			public class linecallstatus
		{
			public uint dwTotalSize;
			public uint dwNeededSize;
			public uint dwUsedSize;
			public LineCallState dwCallState;
			public uint dwCallStateMode;
			public LineCallPrivilege dwCallPrivilege;
			public LineCallFeature dwCallFeatures;
			public uint dwDevSpecificSize;
			public uint dwDevSpecificOffset;
			public LineCallFeature2 dwCallFeatures2;
			SystemTime tStateEntryTime;
		}

		[StructLayout(LayoutKind.Sequential)]
			public class LineCardEntry
		{
			public int PermanentCardID;
			public int CardNameSize;
			public int CardNameOffset;
			public int CardNumberDigits;                             // TAPI v1.4
			public int SameAreaRuleSize;                             // TAPI v1.4
			public int SameAreaRuleOffset;                           // TAPI v1.4
			public int LongDistanceRuleSize;                         // TAPI v1.4
			public int LongDistanceRuleOffset;                       // TAPI v1.4
			public int InternationalRuleSize;                        // TAPI v1.4
			public int InternationalRuleOffset;                      // TAPI v1.4
			public int Options; 
		}

		[StructLayout(LayoutKind.Sequential)]		
			public struct linecalltreatmententry
		{
			public LineCallTreatment dwCallTreatmentID;
			public uint dwCallTreatmentNameSize;
			public uint dwCallTreatmentNameOffset;
		}

		[StructLayout(LayoutKind.Sequential)]
			public class linecountrylist
		{
			public uint TotalSize;
			public uint NeededSize;       
			public uint UsedSize;         
			public uint NumCountries;     
			public uint CountryListSize;  
			public uint CountryListOffset;
		}

		[StructLayout(LayoutKind.Sequential)]
			public class linecountryentry
		{
			public int CountryID;
			public int CountryCode;  
			public int NextCountryID;  
			public int CountryNameSize;  
			public int CountryNameOffset;  
			public int SameAreaRuleSize;  
			public int SameAreaRuleOffset;  
			public int LongDistanceRuleSize;  
			public int LongDistanceRuleOffset;  
			public int InternationalRuleSize;  
			public int InternationalRuleOffset;
		}
		/// <summary>
		/// Device Capabilities Structure
		/// </summary>
		[StructLayout(LayoutKind.Sequential)]		
			public class linedevcaps
		{
			public uint  dwTotalSize;
			public uint  dwNeededSize;
			public uint  dwUsedSize;

			public uint  dwProviderInfoSize;
			public uint  dwProviderInfoOffset;

			public uint  dwSwitchInfoSize;
			public uint  dwSwitchInfoOffset;

			public uint  dwPermanentLineID;
			public uint  dwLineNameSize;
			public uint  dwLineNameOffset;
			public uint  dwStringFormat; 
			public uint  dwAddressModes;
			public uint  dwNumAddresses;
			public uint  dwBearerModes;
			public uint  dwMaxRate;
			public uint  dwMediaModes;

			public uint  dwGenerateToneModes;
			public uint  dwGenerateToneMaxNumFreq;
			public uint  dwGenerateDigitModes;
			public uint  dwMonitorToneMaxNumFreq;
			public uint  dwMonitorToneMaxNumEntries;
			public uint  dwMonitorDigitModes;
			public uint  dwGatherDigitsMinTimeout;
			public uint  dwGatherDigitsMaxTimeout;

			public uint  dwMedCtlDigitMaxListSize;
			public uint  dwMedCtlMediaMaxListSize;
			public uint  dwMedCtlToneMaxListSize;
			public uint  dwMedCtlCallStateMaxListSize;

			public uint  dwDevCapFlags;
			public uint  dwMaxNumActiveCalls;
			public LineAnswerMode  dwAnswerMode;
			public uint  dwRingModes;
			public uint  dwLineStates;

			public uint  dwUUIAcceptSize;
			public uint  dwUUIAnswerSize;
			public uint  dwUUIMakeCallSize;
			public uint  dwUUIDropSize;
			public uint  dwUUISendUserUserInfoSize;
			public uint  dwUUICallInfoSize;

			public linedialparams  MinDialParams;
			public linedialparams  MaxDialParams;
			public linedialparams  DefaultDialParams;

			public uint  dwNumTerminals;
			public uint  dwTerminalCapsSize;
			public uint  dwTerminalCapsOffset;
			public uint  dwTerminalTextEntrySize;
			public uint  dwTerminalTextSize;
			public uint  dwTerminalTextOffset;

			public uint  dwDevSpecificSize;
			public uint  dwDevSpecificOffset;

			public uint  dwLineFeatures;

			public uint  dwSettableDevStatus;
			public uint  dwDeviceClassesSize;
			public uint  dwDeviceClassesOffset;
#if TAPI22
			public System.Guid PermanentLineGuid;      //TAPI Version 2.2
#endif
			/*
			[MarshalAs(UnmanagedType.ByValArray, SizeConst=1 )] 
			public byte[] buffer; // to handle variable-length struct
			*/
		}

	
		/// <summary>
		/// dial params as part of device capabilities
		/// </summary>
		[StructLayout(LayoutKind.Sequential)]
		public struct linedialparams
		{
			public uint dwDialPause;
			public uint dwDialSpeed;
			public uint dwDigitDuration;
			public uint dwWaitForDialtone;
		}

		[StructLayout(LayoutKind.Sequential)]
		public struct lineextensionid 
		{
			public uint  dwExtensionID0;
			public uint  dwExtensionID1;
			public uint  dwExtensionID2;
			public uint  dwExtensionID3;
		}
  
		[StructLayout(LayoutKind.Sequential)]
		public struct linegeneratetone 
		{
			public uint  dwFrequency;
			public uint  dwCadenceOn;
			public uint  dwCadenceOff;
			public uint  dwVolume;
		}
      				
	
		/// <summary>
		/// line initialize extended parameters for LineInitializeEx
		/// </summary>
		[StructLayout(LayoutKind.Sequential)]
		public struct lineinitializeexparams
		{
			public uint  dwTotalSize;
			public uint  dwNeededSize;
			public uint  dwUsedSize;
			public uint  dwOptions;
			public System.IntPtr hEvent;
			public uint  dwCompletionKey;
		}

		[StructLayout(LayoutKind.Sequential)]
			public class LineLocationEntry
		{
			public int PermanentLocationID;
			public int LocationNameSize;
			public int LocationNameOffset;
			public int CountryCode;
			public int CityCodeSize;
			public int CityCodeOffset;
			public int PreferredCardID;
			public int LocalAccessCodeSize;                          // TAPI v1.4
			public int LocalAccessCodeOffset;                        // TAPI v1.4
			public int LongDistanceAccessCodeSize;                   // TAPI v1.4
			public int LongDistanceAccessCodeOffset;                 // TAPI v1.4
			public int TollPrefixListSize;                           // TAPI v1.4
			public int TollPrefixListOffset;                         // TAPI v1.4
			public int CountryID;                                    // TAPI v1.4
			public int Options;                                      // TAPI v1.4
			public int CancelCallWaitingSize;                        // TAPI v1.4
			public int CancelCallWaitingOffset;
		}
		[StructLayout(LayoutKind.Sequential)]
			public struct linemonitortone
		{
			public uint  dwAppSpecific;
			public uint  dwDuration;
			public uint  dwFrequency1;
			public uint  dwFrequency2;
			public uint  dwFrequency3;
		}

		[StructLayout(LayoutKind.Sequential)]         
			public struct linetermcaps
		{
			public uint  dwTermDev;
			public uint  dwTermModes;
			public uint  dwTermSharing;
		}

		[StructLayout(LayoutKind.Sequential)]         
			public class linetranslateoutput
		{
			public uint  dwTotalSize;
			public uint  dwNeededSize;
			public uint  dwUsedSize;
			public uint  dwDialableStringSize;
			public uint  dwDialableStringOffset;
			public uint  dwDisplayableStringSize;
			public uint  dwDisplayableStringOffset;
			public uint  dwCurrentCountry;
			public uint  dwDestCountry;
			public LineTranslateResult  dwTranslateResults;
		}

		[StructLayout(LayoutKind.Sequential)]
			public class linetranslationcaps
		{
			public uint TotalSize;
			public uint NeededSize;
			public uint UsedSize;
			public uint NumLocations;
			public uint LocationListSize;
			public uint LocationListOffset;
			public uint CurrentLocationID;
			public uint NumCards;
			public uint CardListSize;
			public uint CardListOffset;
			public uint CurrentPreferredCardID;
		}
		[StructLayout(LayoutKind.Sequential)]
			public struct SystemTime
		{
			public System.UInt16 wYear;
			public System.UInt16 wMonth;
			public System.UInt16 wDayOfWeek;
			public System.UInt16 wDay;
			public System.UInt16 wHour;
			public System.UInt16 wMinute;
			public System.UInt16 wSecond;
			public System.UInt16 wMilliseconds;
		}

		[StructLayout(LayoutKind.Sequential)]
			public struct varstring
		{
			public uint dwTotalSize;
			public uint dwNeededSize;
			public uint dwUsedSize;
			public uint dwStringFormat;
			public uint dwStringSize;
			public uint dwStringOffset;
			[MarshalAs(UnmanagedType.ByValArray, SizeConst=1024)] 
			public byte[] buffer; // to handle variable-length struct
		}  

		#endregion //structs & helper classes
					
		#region Enumerations
  
		[Flags]
		public enum LineAddrCapFlags: uint
		{
			LINEADDRCAPFLAGS_FWDNUMRINGS = (uint)0x00000001,
			LINEADDRCAPFLAGS_PICKUPGROUPID = (uint)0x00000002,
			LINEADDRCAPFLAGS_SECURE = (uint)0x00000004,
			LINEADDRCAPFLAGS_BLOCKIDDEFAULT = (uint)0x00000008,
			LINEADDRCAPFLAGS_BLOCKIDOVERRIDE = (uint)0x00000010,
			LINEADDRCAPFLAGS_DIALED = (uint)0x00000020,
			LINEADDRCAPFLAGS_ORIGOFFHOOK = (uint)0x00000040,
			LINEADDRCAPFLAGS_DESTOFFHOOK = (uint)0x00000080,
			LINEADDRCAPFLAGS_FWDCONSULT = (uint)0x00000100,
			LINEADDRCAPFLAGS_SETUPCONFNULL = (uint)0x00000200,
			LINEADDRCAPFLAGS_AUTORECONNECT = (uint)0x00000400,
			LINEADDRCAPFLAGS_COMPLETIONID = (uint)0x00000800,
			LINEADDRCAPFLAGS_TRANSFERHELD = (uint)0x00001000,
			LINEADDRCAPFLAGS_TRANSFERMAKE = (uint)0x00002000,
			LINEADDRCAPFLAGS_CONFERENCEHELD = (uint)0x00004000,
			LINEADDRCAPFLAGS_CONFERENCEMAKE = (uint)0x00008000,
			LINEADDRCAPFLAGS_PARTIALDIAL = (uint)0x00010000,
			LINEADDRCAPFLAGS_FWDSTATUSVALID = (uint)0x00020000,
			LINEADDRCAPFLAGS_FWDINTEXTADDR = (uint)0x00040000,
			LINEADDRCAPFLAGS_FWDBUSYNAADDR = (uint)0x00080000,
			LINEADDRCAPFLAGS_ACCEPTTOALERT = (uint)0x00100000,
			LINEADDRCAPFLAGS_CONFDROP = (uint)0x00200000,
			LINEADDRCAPFLAGS_PICKUPCALLWAIT = (uint)0x00400000,
			LINEADDRCAPFLAGS_PREDICTIVEDIALER = (uint)0x00800000,       // TAPI v2.0
			LINEADDRCAPFLAGS_QUEUE = (uint)0x01000000,                  // TAPI v2.0
			LINEADDRCAPFLAGS_ROUTEPOINT = (uint)0x02000000,             // TAPI v2.0
			LINEADDRCAPFLAGS_HOLDMAKESNEW = (uint)0x04000000,           // TAPI v2.0
			LINEADDRCAPFLAGS_NOINTERNALCALLS = (uint)0x08000000,        // TAPI v2.0
			LINEADDRCAPFLAGS_NOEXTERNALCALLS = (uint)0x10000000,        // TAPI v2.0
			LINEADDRCAPFLAGS_SETCALLINGID = (uint)0x20000000            // TAPI v2.0
		}
  
		public enum LineAddressMode: uint
		{
			LINEADDRESSMODE_ADDRESSID = (uint)0x00000001,
			LINEADDRESSMODE_DIALABLEADDR = (uint)0x00000002
		}
  
		[Flags]
			public enum LineAddressSharing: uint
		{
			LINEADDRESSSHARING_PRIVATE = (uint)0x00000001,
			LINEADDRESSSHARING_BRIDGEDEXCL = (uint)0x00000002,
			LINEADDRESSSHARING_BRIDGEDNEW = (uint)0x00000004,
			LINEADDRESSSHARING_BRIDGEDSHARED = (uint)0x00000008,
			LINEADDRESSSHARING_MONITORED = (uint)0x00000010    
		}
  
		[Flags]
			public enum LineAddressState: uint
		{
			LINEADDRESSSTATE_OTHER = (uint)0x00000001,
			LINEADDRESSSTATE_DEVSPECIFIC = (uint)0x00000002,
			LINEADDRESSSTATE_INUSEZERO = (uint)0x00000004,
			LINEADDRESSSTATE_INUSEONE = (uint)0x00000008,
			LINEADDRESSSTATE_INUSEMANY = (uint)0x00000010,
			LINEADDRESSSTATE_NUMCALLS = (uint)0x00000020,
			LINEADDRESSSTATE_FORWARD = (uint)0x00000040,
			LINEADDRESSSTATE_TERMINALS = (uint)0x00000080,
			LINEADDRESSSTATE_CAPSCHANGE = (uint)0x00000100      // TAPI v1.4    
		}
  
		[Flags]
			public enum LineAddrFeature: uint
		{
			LINEADDRFEATURE_FORWARD = (uint)0x00000001,
			LINEADDRFEATURE_MAKECALL = (uint)0x00000002,
			LINEADDRFEATURE_PICKUP = (uint)0x00000004,
			LINEADDRFEATURE_SETMEDIACONTROL = (uint)0x00000008,
			LINEADDRFEATURE_SETTERMINAL = (uint)0x00000010,
			LINEADDRFEATURE_SETUPCONF = (uint)0x00000020,
			LINEADDRFEATURE_UNCOMPLETECALL = (uint)0x00000040,
			LINEADDRFEATURE_UNPARK = (uint)0x00000080,
			LINEADDRFEATURE_PICKUPHELD = (uint)0x00000100,          // TAPI v2.0
			LINEADDRFEATURE_PICKUPGROUP = (uint)0x00000200,         // TAPI v2.0
			LINEADDRFEATURE_PICKUPDIRECT = (uint)0x00000400,        // TAPI v2.0
			LINEADDRFEATURE_PICKUPWAITING = (uint)0x00000800,       // TAPI v2.0
			LINEADDRFEATURE_FORWARDFWD = (uint)0x00001000,          // TAPI v2.0
			LINEADDRFEATURE_FORWARDDND = (uint)0x00002000           // TAPI v2.0
		}
  
		[Flags]
			public enum LineAnswerMode: uint
		{
			LINEANSWERMODE_NONE = (uint)0x00000001,
			LINEANSWERMODE_DROP = (uint)0x00000002,
			LINEANSWERMODE_HOLD = (uint)0x00000004    
		}
              
		[Flags]
			public enum LineBearerMode: uint
		{
			LINEBEARERMODE_VOICE = (uint)0x00000001,
			LINEBEARERMODE_SPEECH = (uint)0x00000002,
			LINEBEARERMODE_MULTIUSE = (uint)0x00000004,
			LINEBEARERMODE_DATA = (uint)0x00000008,
			LINEBEARERMODE_ALTSPEECHDATA = (uint)0x00000010,
			LINEBEARERMODE_NONCALLSIGNALING = (uint)0x00000020,
			LINEBEARERMODE_PASSTHROUGH = (uint)0x00000040,       // TAPI v1.4
			LINEBEARERMODE_RESTRICTEDDATA = (uint)0x00000080     // TAPI v2.0    
		}

		[Flags]
			public enum LineBusyMode: uint
		{
			LINEBUSYMODE_STATION = (uint)0x00000001,
			LINEBUSYMODE_TRUNK = (uint)0x00000002,
			LINEBUSYMODE_UNKNOWN = (uint)0x00000004,
			LINEBUSYMODE_UNAVAIL = (uint)0x00000008    
		}
   
		[Flags] 
			public enum LineCallComplMode: uint
		{
			LINECALLCOMPLCOND_BUSY = (uint)0x00000001,
			LINECALLCOMPLCOND_NOANSWER = (uint)0x00000002    
		}
  
		[Flags]
			public enum LineCallFeature: uint
		{
			LINECALLFEATURE_ACCEPT = (uint)0x00000001,
			LINECALLFEATURE_ADDTOCONF = (uint)0x00000002,
			LINECALLFEATURE_ANSWER = (uint)0x00000004,
			LINECALLFEATURE_BLINDTRANSFER = (uint)0x00000008,
			LINECALLFEATURE_COMPLETECALL = (uint)0x00000010,
			LINECALLFEATURE_COMPLETETRANSF = (uint)0x00000020,
			LINECALLFEATURE_DIAL = (uint)0x00000040,
			LINECALLFEATURE_DROP = (uint)0x00000080,
			LINECALLFEATURE_GATHERDIGITS = (uint)0x00000100,
			LINECALLFEATURE_GENERATEDIGITS = (uint)0x00000200,
			LINECALLFEATURE_GENERATETONE = (uint)0x00000400,
			LINECALLFEATURE_HOLD = (uint)0x00000800,
			LINECALLFEATURE_MONITORDIGITS = (uint)0x00001000,
			LINECALLFEATURE_MONITORMEDIA = (uint)0x00002000,
			LINECALLFEATURE_MONITORTONES = (uint)0x00004000,
			LINECALLFEATURE_PARK = (uint)0x00008000,
			LINECALLFEATURE_PREPAREADDCONF = (uint)0x00010000,
			LINECALLFEATURE_REDIRECT = (uint)0x00020000,
			LINECALLFEATURE_REMOVEFROMCONF = (uint)0x00040000,
			LINECALLFEATURE_SECURECALL = (uint)0x00080000,
			LINECALLFEATURE_SENDUSERUSER = (uint)0x00100000,
			LINECALLFEATURE_SETCALLPARAMS = (uint)0x00200000,
			LINECALLFEATURE_SETMEDIACONTROL = (uint)0x00400000,
			LINECALLFEATURE_SETTERMINAL = (uint)0x00800000,
			LINECALLFEATURE_SETUPCONF = (uint)0x01000000,
			LINECALLFEATURE_SETUPTRANSFER = (uint)0x02000000,
			LINECALLFEATURE_SWAPHOLD = (uint)0x04000000,
			LINECALLFEATURE_UNHOLD = (uint)0x08000000,
			LINECALLFEATURE_RELEASEUSERUSERINFO = (uint)0x10000000,      // TAPI v1.4
			LINECALLFEATURE_SETTREATMENT = (uint)0x20000000,      // TAPI v2.0
			LINECALLFEATURE_SETQOS = (uint)0x40000000,      // TAPI v2.0
			LINECALLFEATURE_SETCALLDATA = (uint)0x80000000,      // TAPI v2.0
		}
  
		[Flags]
			public enum LineCallFeature2: uint
		{
			LINECALLFEATURE2_NOHOLDCONFERENCE = (uint)0x00000001,       // TAPI v2.0
			LINECALLFEATURE2_ONESTEPTRANSFER = (uint)0x00000002,        // TAPI v2.0
			LINECALLFEATURE2_COMPLCAMPON = (uint)0x00000004,            // TAPI v2.0
			LINECALLFEATURE2_COMPLCALLBACK = (uint)0x00000008,          // TAPI v2.0
			LINECALLFEATURE2_COMPLINTRUDE = (uint)0x00000010,           // TAPI v2.0
			LINECALLFEATURE2_COMPLMESSAGE = (uint)0x00000020,           // TAPI v2.0
			LINECALLFEATURE2_TRANSFERNORM = (uint)0x00000040,           // TAPI v2.0
			LINECALLFEATURE2_TRANSFERCONF = (uint)0x00000080,           // TAPI v2.0
			LINECALLFEATURE2_PARKDIRECT = (uint)0x00000100,             // TAPI v2.0
			LINECALLFEATURE2_PARKNONDIRECT = (uint)0x00000200           // TAPI v2.0
		}
      
		[Flags]
			public enum LineCallInfoState: uint
		{
			LINECALLINFOSTATE_OTHER = (uint)0x00000001,
			LINECALLINFOSTATE_DEVSPECIFIC = (uint)0x00000002,
			LINECALLINFOSTATE_BEARERMODE = (uint)0x00000004,
			LINECALLINFOSTATE_RATE = (uint)0x00000008,
			LINECALLINFOSTATE_MEDIAMODE = (uint)0x00000010,
			LINECALLINFOSTATE_APPSPECIFIC = (uint)0x00000020,
			LINECALLINFOSTATE_CALLID = (uint)0x00000040,
			LINECALLINFOSTATE_RELATEDCALLID = (uint)0x00000080,
			LINECALLINFOSTATE_ORIGIN = (uint)0x00000100,
			LINECALLINFOSTATE_REASON = (uint)0x00000200,
			LINECALLINFOSTATE_COMPLETIONID = (uint)0x00000400,
			LINECALLINFOSTATE_NUMOWNERINCR = (uint)0x00000800,
			LINECALLINFOSTATE_NUMOWNERDECR = (uint)0x00001000,
			LINECALLINFOSTATE_NUMMONITORS = (uint)0x00002000,
			LINECALLINFOSTATE_TRUNK = (uint)0x00004000,
			LINECALLINFOSTATE_CALLERID = (uint)0x00008000,
			LINECALLINFOSTATE_CALLEDID = (uint)0x00010000,
			LINECALLINFOSTATE_CONNECTEDID = (uint)0x00020000,
			LINECALLINFOSTATE_REDIRECTIONID = (uint)0x00040000,
			LINECALLINFOSTATE_REDIRECTINGID = (uint)0x00080000,
			LINECALLINFOSTATE_DISPLAY = (uint)0x00100000,
			LINECALLINFOSTATE_USERUSERINFO = (uint)0x00200000,
			LINECALLINFOSTATE_HIGHLEVELCOMP = (uint)0x00400000,
			LINECALLINFOSTATE_LOWLEVELCOMP = (uint)0x00800000,
			LINECALLINFOSTATE_CHARGINGINFO = (uint)0x01000000,
			LINECALLINFOSTATE_TERMINAL = (uint)0x02000000,
			LINECALLINFOSTATE_DIALPARAMS = (uint)0x04000000,
			LINECALLINFOSTATE_MONITORMODES = (uint)0x08000000,
			LINECALLINFOSTATE_TREATMENT = (uint)0x10000000,      // TAPI v2.0
			LINECALLINFOSTATE_QOS = (uint)0x20000000,            // TAPI v2.0
			LINECALLINFOSTATE_CALLDATA = (uint)0x40000000        // TAPI v2.0
		}
  
		public enum LineCallOrigin: uint
		{
			LINECALLORIGIN_OUTBOUND = (uint)0x00000001,
			LINECALLORIGIN_INTERNAL = (uint)0x00000002,
			LINECALLORIGIN_EXTERNAL = (uint)0x00000004,
			LINECALLORIGIN_UNKNOWN = (uint)0x00000010,
			LINECALLORIGIN_UNAVAIL = (uint)0x00000020,
			LINECALLORIGIN_CONFERENCE = (uint) 0x00000040,
			LINECALLORIGIN_INBOUND = (uint)0x00000080      // TAPI v1.4    
		}
  
		[Flags]
			public enum LineCallParamFlags: uint
		{
			LINECALLPARAMFLAGS_SECURE = (uint)0x00000001,
			LINECALLPARAMFLAGS_IDLE = (uint)0x00000002,
			LINECALLPARAMFLAGS_BLOCKID = (uint)0x00000004,
			LINECALLPARAMFLAGS_ORIGOFFHOOK = (uint)0x00000008,
			LINECALLPARAMFLAGS_DESTOFFHOOK = (uint)0x00000010,
			LINECALLPARAMFLAGS_NOHOLDCONFERENCE = (uint)0x00000020,       // TAPI v2.0
			LINECALLPARAMFLAGS_PREDICTIVEDIAL = (uint)0x00000040,         // TAPI v2.0
			LINECALLPARAMFLAGS_ONESTEPTRANSFER = (uint)0x00000080         // TAPI v2.0    
		}
  
		[Flags]
			public enum LineCallPartyID: uint
		{
			LINECALLPARTYID_BLOCKED = (uint)0x00000001,
			LINECALLPARTYID_OUTOFAREA = (uint)0x00000002,
			LINECALLPARTYID_NAME = (uint)0x00000004,
			LINECALLPARTYID_ADDRESS = (uint)0x00000008,
			LINECALLPARTYID_PARTIAL = (uint)0x00000010,
			LINECALLPARTYID_UNKNOWN = (uint)0x00000020,
			LINECALLPARTYID_UNAVAIL = (uint)0x00000040
		}
  
		[Flags]
			public enum LineCallPrivilege: uint
		{
			LINECALLPRIVILEGE_NONE = (uint)0x00000001,
			LINECALLPRIVILEGE_MONITOR = (uint)0x00000002,
			LINECALLPRIVILEGE_OWNER = (uint)0x00000004
		}
  
		public enum LineCallReason: uint
		{
			LINECALLREASON_DIRECT = (uint)0x00000001,
			LINECALLREASON_FWDBUSY = (uint)0x00000002,
			LINECALLREASON_FWDNOANSWER = (uint)0x00000004,
			LINECALLREASON_FWDUNCOND = (uint)0x00000008,
			LINECALLREASON_PICKUP = (uint)0x00000010,
			LINECALLREASON_UNPARK = (uint)0x00000020,
			LINECALLREASON_REDIRECT = (uint)0x00000040,
			LINECALLREASON_CALLCOMPLETION = (uint)0x00000080,
			LINECALLREASON_TRANSFER = (uint)0x00000100,
			LINECALLREASON_REMINDER = (uint)0x00000200,
			LINECALLREASON_UNKNOWN = (uint)0x00000400,
			LINECALLREASON_UNAVAIL = (uint)0x00000800,
			LINECALLREASON_INTRUDE = (uint)0x00001000,          // TAPI v1.4
			LINECALLREASON_PARKED = (uint)0x00002000,           // TAPI v1.4
			LINECALLREASON_CAMPEDON = (uint)0x00004000,         // TAPI v2.0
			LINECALLREASON_ROUTEREQUEST = (uint)0x00008000      // TAPI v2.0    
		}
  
		[Flags]
			public enum LineCallState: uint
		{
			LINECALLSTATE_IDLE = (uint)0x00000001,
			LINECALLSTATE_OFFERING = (uint)0x00000002,
			LINECALLSTATE_ACCEPTED = (uint)0x00000004,
			LINECALLSTATE_DIALTONE = (uint)0x00000008,
			LINECALLSTATE_DIALING = (uint)0x00000010,
			LINECALLSTATE_RINGBACK = (uint)0x00000020,
			LINECALLSTATE_BUSY = (uint)0x00000040,
			LINECALLSTATE_SPECIALINFO = (uint)0x00000080,
			LINECALLSTATE_CONNECTED = (uint)0x00000100,
			LINECALLSTATE_PROCEEDING = (uint)0x00000200,
			LINECALLSTATE_ONHOLD = (uint)0x00000400,
			LINECALLSTATE_CONFERENCED = (uint)0x00000800,
			LINECALLSTATE_ONHOLDPENDCONF = (uint)0x00001000,
			LINECALLSTATE_ONHOLDPENDTRANSFER = (uint)0x00002000,
			LINECALLSTATE_DISCONNECTED = (uint)0x00004000,
			LINECALLSTATE_UNKNOWN = (uint)0x00008000
		}

		public enum LineCallTreatment: uint
		{
			LINECALLTREATMENT_SILENCE = (uint)0x00000001,       // TAPI v2.0
			LINECALLTREATMENT_RINGBACK = (uint)0x00000002,      // TAPI v2.0
			LINECALLTREATMENT_BUSY = (uint)0x00000003,          // TAPI v2.0
			LINECALLTREATMENT_MUSIC = (uint)0x00000004          // TAPI v2.0    
		}

		/// <summary>
		/// define values used in the Options member of the 
		/// LINECARDENTRY structure returned as part of the 
		/// LINETRANSLATECAPS structure returned by lineGetTranslateCaps. 
		/// </summary>
		public enum LineCardOption: uint
		{
			/// <summary>
			/// This calling card is one of the predefined calling card 
			/// definitions included by Microsoft with Win32 Telephony. 
			/// It cannot be removed entirely using Dial Helper; if the 
			/// user attempts to remove it, it will become HIDDEN. 
			/// It thus continues to be accessible for copying of dialing rules. 
			/// </summary>
			PREDEFINED              = (uint)0x00000001,      // TAPI v1.4
			/// <summary>
			/// This calling card has been hidden by the user. It is not 
			/// shown by Dial Helper in the main listing of available 
			/// calling cards, but will be shown in the list of cards 
			/// from which dialing rules can be copied.
			/// </summary>
			HIDDEN                  = (uint)0x00000002,      // TAPI v1.4
		}
  
		[Flags]
			public enum LineConnectedMode: uint
		{
			LINECONNECTEDMODE_ACTIVE = (uint)0x00000001,         // TAPI v1.4
			LINECONNECTEDMODE_INACTIVE = (uint)0x00000002,       // TAPI v1.4
			LINECONNECTEDMODE_ACTIVEHELD = (uint)0x00000004,     // TAPI v2.0
			LINECONNECTEDMODE_INACTIVEHELD = (uint)0x00000008,   // TAPI v2.0
			LINECONNECTEDMODE_CONFIRMED = (uint)0x00000010       // TAPI v2.0
		}
     
		[Flags]
			public enum LineDevCapFlags: uint
		{
			LINEDEVCAPFLAGS_CROSSADDRCONF = (uint)0x00000001,
			LINEDEVCAPFLAGS_HIGHLEVCOMP = (uint)0x00000002,
			LINEDEVCAPFLAGS_LOWLEVCOMP = (uint)0x00000004,
			LINEDEVCAPFLAGS_MEDIACONTROL = (uint)0x00000008,
			LINEDEVCAPFLAGS_MULTIPLEADDR = (uint)0x00000010,
			LINEDEVCAPFLAGS_CLOSEDROP = (uint)0x00000020,
			LINEDEVCAPFLAGS_DIALBILLING = (uint)0x00000040,
			LINEDEVCAPFLAGS_DIALQUIET = (uint)0x00000080,
			LINEDEVCAPFLAGS_DIALDIALTONE = (uint)0x00000100    
		}
  
		[Flags]
			public enum LineDevState: uint
		{
			LINEDEVSTATE_OTHER = (uint)0x00000001,
			LINEDEVSTATE_RINGING = (uint)0x00000002,
			LINEDEVSTATE_CONNECTED = (uint)0x00000004,
			LINEDEVSTATE_DISCONNECTED = (uint)0x00000008,
			LINEDEVSTATE_MSGWAITON = (uint)0x00000010,
			LINEDEVSTATE_MSGWAITOFF = (uint)0x00000020,
			LINEDEVSTATE_INSERVICE = (uint)0x00000040,
			LINEDEVSTATE_OUTOFSERVICE = (uint)0x00000080,
			LINEDEVSTATE_MAINTENANCE = (uint)0x00000100,
			LINEDEVSTATE_OPEN = (uint)0x00000200,
			LINEDEVSTATE_CLOSE = (uint)0x00000400,
			LINEDEVSTATE_NUMCALLS = (uint)0x00000800,
			LINEDEVSTATE_NUMCOMPLETIONS = (uint)0x00001000,
			LINEDEVSTATE_TERMINALS = (uint)0x00002000,
			LINEDEVSTATE_ROAMMODE = (uint)0x00004000,
			LINEDEVSTATE_BATTERY = (uint)0x00008000,
			LINEDEVSTATE_SIGNAL = (uint)0x00010000,
			LINEDEVSTATE_DEVSPECIFIC = (uint)0x00020000,
			LINEDEVSTATE_REINIT = (uint)0x00040000,
			LINEDEVSTATE_LOCK = (uint)0x00080000,
			LINEDEVSTATE_CAPSCHANGE = (uint)0x00100000,           // TAPI v1.4
			LINEDEVSTATE_CONFIGCHANGE = (uint)0x00200000,         // TAPI v1.4
			LINEDEVSTATE_TRANSLATECHANGE = (uint)0x00400000,      // TAPI v1.4
			LINEDEVSTATE_COMPLCANCEL = (uint)0x00800000,          // TAPI v1.4
			LINEDEVSTATE_REMOVED = (uint)0x01000000               // TAPI v1.4    
		}

   
		[Flags]
			public enum LineDevStatus: uint
		{
			LINEDEVSTATUSFLAGS_CONNECTED = (uint)0x00000001,
			LINEDEVSTATUSFLAGS_MSGWAIT = (uint)0x00000002,
			LINEDEVSTATUSFLAGS_INSERVICE = (uint)0x00000004,
			LINEDEVSTATUSFLAGS_LOCKED = (uint)0x00000008    
		}
   
		[Flags]
			public enum LineDialToneMode: uint
		{
			LINEDIALTONEMODE_NORMAL = (uint)0x00000001,
			LINEDIALTONEMODE_SPECIAL = (uint)0x00000002,
			LINEDIALTONEMODE_INTERNAL = (uint)0x00000004,
			LINEDIALTONEMODE_EXTERNAL = (uint)0x00000008,
			LINEDIALTONEMODE_UNKNOWN = (uint)0x00000010,
			LINEDIALTONEMODE_UNAVAIL = (uint)0x00000020    
		}
                
		public enum LineDigitMode: uint
		{
			LINEDIGITMODE_PULSE = (uint)0x00000001,
			LINEDIGITMODE_DTMF = (uint)0x00000002,
			LINEDIGITMODE_DTMFEND = (uint)0x00000004    
		}
  
		[Flags]
			public enum LineDisconnectedMode: uint
		{
			LINEDISCONNECTMODE_NORMAL = (uint)0x00000001,
			LINEDISCONNECTMODE_UNKNOWN = (uint)0x00000002,
			LINEDISCONNECTMODE_REJECT = (uint)0x00000004,
			LINEDISCONNECTMODE_PICKUP = (uint)0x00000008,
			LINEDISCONNECTMODE_FORWARDED = (uint)0x00000010,
			LINEDISCONNECTMODE_BUSY = (uint)0x00000020,
			LINEDISCONNECTMODE_NOANSWER = (uint)0x00000040,
			LINEDISCONNECTMODE_BADADDRESS = (uint)0x00000080,
			LINEDISCONNECTMODE_UNREACHABLE = (uint)0x00000100,
			LINEDISCONNECTMODE_CONGESTION = (uint)0x00000200,
			LINEDISCONNECTMODE_INCOMPATIBLE = (uint)0x00000400,
			LINEDISCONNECTMODE_UNAVAIL = (uint) 0x00000800,
			LINEDISCONNECTMODE_NODIALTONE = (uint)0x00001000,         // TAPI v1.4
			LINEDISCONNECTMODE_NUMBERCHANGED = (uint)0x00002000,      // TAPI v2.0
			LINEDISCONNECTMODE_OUTOFORDER = (uint)0x00004000,         // TAPI v2.0
			LINEDISCONNECTMODE_TEMPFAILURE = (uint)0x00008000,        // TAPI v2.0
			LINEDISCONNECTMODE_QOSUNAVAIL = (uint)0x00010000,         // TAPI v2.0
			LINEDISCONNECTMODE_BLOCKED = (uint)0x00020000,            // TAPI v2.0
			LINEDISCONNECTMODE_DONOTDISTURB = (uint)0x00040000,       // TAPI v2.0
			LINEDISCONNECTMODE_CANCELLED = (uint)0x00080000           // TAPI v2.0
		}
      		
		/// <summary>
		/// returns from lineInitializeEx, lineInitialize, lineShutDown, lineNegotiateAPIVersion
		/// </summary>
		public enum LineErrReturn: uint
		{
			LINEERR_OK = (uint)0x00000000,
			LINEERR_ALLOCATED = (uint)0x80000001,
			LINEERR_BADDEVICEID = (uint)0x80000002,
			LINEERR_BEARERMODEUNAVAIL = (uint)0x80000003,
			LINEERR_CALLUNAVAIL = (uint)0x80000005,
			LINEERR_COMPLETIONOVERRUN = (uint)0x80000006,
			LINEERR_CONFERENCEFULL = (uint)0x80000007,
			LINEERR_DIALBILLING = (uint)0x80000008,
			LINEERR_DIALDIALTONE = (uint)0x80000009,
			LINEERR_DIALPROMPT = (uint)0x8000000A,
			LINEERR_DIALQUIET = (uint)0x8000000B,
			LINEERR_INCOMPATIBLEAPIVERSION = (uint)0x8000000C,
			LINEERR_INCOMPATIBLEEXTVERSION = (uint)0x8000000D,
			LINEERR_INIFILECORRUPT = (uint)0x8000000E,
			LINEERR_INUSE = (uint)0x8000000F,
			LINEERR_INVALADDRESS = (uint)0x80000010,
			LINEERR_INVALADDRESSID = (uint)0x80000011,
			LINEERR_INVALADDRESSMODE = (uint)0x80000012,
			LINEERR_INVALADDRESSSTATE = (uint)0x80000013,
			LINEERR_INVALAPPHANDLE = (uint)0x80000014,
			LINEERR_INVALAPPNAME = (uint)0x80000015,
			LINEERR_INVALBEARERMODE = (uint)0x80000016,
			LINEERR_INVALCALLCOMPLMODE = (uint)0x80000017,
			LINEERR_INVALCALLHANDLE = (uint)0x80000018,
			LINEERR_INVALCALLPARAMS = (uint)0x80000019,
			LINEERR_INVALCALLPRIVILEGE = (uint)0x8000001A,
			LINEERR_INVALCALLSELECT = (uint)0x8000001B,
			LINEERR_INVALCALLSTATE = (uint)0x8000001C,
			LINEERR_INVALCALLSTATELIST = (uint)0x8000001D,
			LINEERR_INVALCARD = (uint)0x8000001E,
			LINEERR_INVALCOMPLETIONID = (uint)0x8000001F,
			LINEERR_INVALCONFCALLHANDLE = (uint)0x80000020,
			LINEERR_INVALCONSULTCALLHANDLE = (uint)0x80000021,
			LINEERR_INVALCOUNTRYCODE = (uint)0x80000022,
			LINEERR_INVALDEVICECLASS = (uint)0x80000023,
			LINEERR_INVALDEVICEHANDLE = (uint)0x80000024,
			LINEERR_INVALDIALPARAMS = (uint)0x80000025,
			LINEERR_INVALDIGITLIST = (uint)0x80000026,
			LINEERR_INVALDIGITMODE = (uint)0x80000027,
			LINEERR_INVALDIGITS = (uint)0x80000028,
			LINEERR_INVALEXTVERSION = (uint)0x80000029,
			LINEERR_INVALGROUPID = (uint)0x8000002A,
			LINEERR_INVALLINEHANDLE = (uint)0x8000002B,
			LINEERR_INVALLINESTATE = (uint)0x8000002C,
			LINEERR_INVALLOCATION = (uint)0x8000002D,
			LINEERR_INVALMEDIALIST = (uint)0x8000002E,
			LINEERR_INVALMEDIAMODE = (uint)0x8000002F,
			LINEERR_INVALMESSAGEID = (uint)0x80000030,
			LINEERR_INVALPARAM = (uint)0x80000032,
			LINEERR_INVALPARKID = (uint)0x80000033,
			LINEERR_INVALPARKMODE = (uint)0x80000034,
			LINEERR_INVALPOINTER = (uint)0x80000035,
			LINEERR_INVALPRIVSELECT = (uint)0x80000036,
			LINEERR_INVALRATE = (uint)0x80000037,
			LINEERR_INVALREQUESTMODE = (uint)0x80000038,
			LINEERR_INVALTERMINALID = (uint)0x80000039,
			LINEERR_INVALTERMINALMODE = (uint)0x8000003A,
			LINEERR_INVALTIMEOUT = (uint)0x8000003B,
			LINEERR_INVALTONE = (uint)0x8000003C,
			LINEERR_INVALTONELIST = (uint)0x8000003D,
			LINEERR_INVALTONEMODE = (uint)0x8000003E,
			LINEERR_INVALTRANSFERMODE = (uint)0x8000003F,
			LINEERR_LINEMAPPERFAILED = (uint)0x80000040,
			LINEERR_NOCONFERENCE = (uint)0x80000041,
			LINEERR_NODEVICE = (uint)0x80000042,
			LINEERR_NODRIVER = (uint)0x80000043,
			LINEERR_NOMEM = (uint)0x80000044,
			LINEERR_NOREQUEST = (uint)0x80000045,
			LINEERR_NOTOWNER = (uint)0x80000046,
			LINEERR_NOTREGISTERED = (uint)0x80000047,
			LINEERR_OPERATIONFAILED = (uint)0x80000048,
			LINEERR_OPERATIONUNAVAIL = (uint)0x80000049,
			LINEERR_RATEUNAVAIL = (uint)0x8000004A,
			LINEERR_RESOURCEUNAVAIL = (uint)0x8000004B,
			LINEERR_REQUESTOVERRUN = (uint)0x8000004C,
			LINEERR_STRUCTURETOOSMALL = (uint)0x8000004D,
			LINEERR_TARGETNOTFOUND = (uint)0x8000004E,
			LINEERR_TARGETSELF = (uint)0x8000004F,
			LINEERR_UNINITIALIZED = (uint)0x80000050,
			LINEERR_USERUSERINFOTOOBIG = (uint)0x80000051,
			LINEERR_REINIT = (uint)0x80000052,
			LINEERR_ADDRESSBLOCKED = (uint)0x80000053,
			LINEERR_BILLINGREJECTED = (uint)0x80000054,
			LINEERR_INVALFEATURE = (uint)0x80000055,
			LINEERR_NOMULTIPLEINSTANCE = (uint)0x80000056,
			LINEERR_INVALAGENTID = (uint)0x80000057,
			LINEERR_INVALAGENTGROUP = (uint)0x80000058,
			LINEERR_INVALPASSWORD = (uint)0x80000059,
			LINEERR_INVALAGENTSTATE = (uint)0x8000005A,
			LINEERR_INVALAGENTACTIVITY = (uint)0x8000005B,
			LINEERR_DIALVOICEDETECT = (uint)0x8000005C
		};
  
		[Flags]
			public enum LineFeature: uint
		{
			LINEFEATURE_DEVSPECIFIC = (uint)0x00000001,
			LINEFEATURE_DEVSPECIFICFEAT = (uint)0x00000002,
			LINEFEATURE_FORWARD = (uint)0x00000004,
			LINEFEATURE_MAKECALL = (uint)0x00000008,
			LINEFEATURE_SETMEDIACONTROL = (uint)0x00000010,
			LINEFEATURE_SETTERMINAL = (uint)0x00000020,
			LINEFEATURE_SETDEVSTATUS = (uint)0x00000040,      // TAPI v2.0
			LINEFEATURE_FORWARDFWD = (uint)0x00000080,        // TAPI v2.0
			LINEFEATURE_FORWARDDND = (uint)0x00000100         // TAPI v2.0    
		}

    [Flags]
    public enum LineForwardMode : uint
		{
			LINEFORWARDMODE_UNCOND = (uint)0x00000001,
			LINEFORWARDMODE_UNCONDINTERNAL = (uint)0x00000002,
			LINEFORWARDMODE_UNCONDEXTERNAL = (uint)0x00000004,
			LINEFORWARDMODE_UNCONDSPECIFIC = (uint)0x00000008,
			LINEFORWARDMODE_BUSY = (uint)0x00000010,
			LINEFORWARDMODE_BUSYINTERNAL = (uint)0x00000020,
			LINEFORWARDMODE_BUSYEXTERNAL = (uint)0x00000040,
			LINEFORWARDMODE_BUSYSPECIFIC = (uint)0x00000080,
			LINEFORWARDMODE_NOANSW = (uint)0x00000100,
			LINEFORWARDMODE_NOANSWINTERNAL = (uint)0x00000200,
			LINEFORWARDMODE_NOANSWEXTERNAL = (uint)0x00000400,
			LINEFORWARDMODE_NOANSWSPECIFIC = (uint) 0x00000800,
			LINEFORWARDMODE_BUSYNA = (uint)0x00001000,
			LINEFORWARDMODE_BUSYNAINTERNAL = (uint)0x00002000,
			LINEFORWARDMODE_BUSYNAEXTERNAL = (uint)0x00004000,
			LINEFORWARDMODE_BUSYNASPECIFIC = (uint)0x00008000,
			LINEFORWARDMODE_UNKNOWN = (uint)0x00010000,           // TAPI v1.4
			LINEFORWARDMODE_UNAVAIL = (uint)0x00020000            // TAPI v1.4    
		}

    [Flags]
    public enum LineGatherTerm: uint
    {
      LINEGATHERTERM_BUFFERFULL = (uint)0x00000001,
      LINEGATHERTERM_TERMDIGIT = (uint)0x00000002,
      LINEGATHERTERM_FIRSTTIMEOUT = (uint)0x00000004,
      LINEGATHERTERM_INTERTIMEOUT = (uint)0x00000008,
      LINEGATHERTERM_CANCEL = (uint)0x0000001
    }
  		
    [Flags]
    public enum LineGenerateTerm: uint
    {
      LINEGENERATETERM_CANCEL = (uint)0x00000002,
      LINEGENERATETERM_DONE  = (uint)0x00000001
    }

		public enum LineInitializeExOptions: uint
		{
			LINEINITIALIZEEXOPTION_USEHIDDENWINDOW = (uint)0x00000001,
			LINEINITIALIZEEXOPTION_USEEVENT = (uint)0x00000002,
			LINEINITIALIZEEXOPTION_USECOMPLETIONPORT = (uint)0x00000003     
		}		

		[Flags]
		public enum LineMediaMode: uint
		{
			LINEMEDIAMODE_UNKNOWN = (uint)0x00000002,
			LINEMEDIAMODE_INTERACTIVEVOICE = (uint)0x00000004,
			LINEMEDIAMODE_AUTOMATEDVOICE = (uint)0x00000008,
			LINEMEDIAMODE_DATAMODEM = (uint)0x00000010,
			LINEMEDIAMODE_G3FAX = (uint)0x00000020,
			LINEMEDIAMODE_TDD = (uint)0x00000040,
			LINEMEDIAMODE_G4FAX = (uint)0x00000080,
			LINEMEDIAMODE_DIGITALDATA = (uint)0x00000100,
			LINEMEDIAMODE_TELETEX = (uint)0x00000200,
			LINEMEDIAMODE_VIDEOTEX = (uint)0x00000400,
			LINEMEDIAMODE_TELEX = (uint)0x00000800,
			LINEMEDIAMODE_MIXED = (uint)0x00001000,
			LINEMEDIAMODE_ADSI = (uint)0x00002000,
			LINEMEDIAMODE_VOICEVIEW = (uint)0x00004000,      
			LINEMEDIAMODE_VIDEO = (uint)0x00008000          
		}
  
		public enum LineOfferingMode: uint
		{
			LINEOFFERINGMODE_ACTIVE = (uint)0x00000001,       // TAPI v1.4
			LINEOFFERINGMODE_INACTIVE = (uint)0x00000002      // TAPI v1.4
		}
  
		public enum LineOpenOption: uint
		{
			LINEOPENOPTION_SINGLEADDRESS = (uint)0x80000000,      // TAPI v2.0
			LINEOPENOPTION_PROXY = (uint)0x40000000               // TAPI v2.0
		}
  
		public enum LineParkMode: uint
		{
			LINEPARKMODE_DIRECTED = (uint)0x00000001,
			LINEPARKMODE_NONDIRECTED = (uint)0x00000002    
		}
  
		public enum LineRemoveFromConf: uint
		{
			LINEREMOVEFROMCONF_NONE = (uint)0x00000001,
			LINEREMOVEFROMCONF_LAST = (uint)0x00000002,
			LINEREMOVEFROMCONF_ANY = (uint)0x00000003
		}
  
		[Flags]
			public enum LineSpecialInfo: uint
		{
			LINESPECIALINFO_NOCIRCUIT = (uint)0x00000001,
			LINESPECIALINFO_CUSTIRREG = (uint)0x00000002,
			LINESPECIALINFO_REORDER = (uint)0x00000004,
			LINESPECIALINFO_UNKNOWN = (uint)0x00000008,
			LINESPECIALINFO_UNAVAIL = (uint)0x00000010
		}
       
		[Flags]
			public enum LineTermDev: uint
		{
			LINETERMDEV_PHONE = (uint)0x00000001,
			LINETERMDEV_HEADSET = (uint)0x00000002,
			LINETERMDEV_SPEAKER = (uint)0x00000004    
		}
   
		[Flags]
			public enum LineTermMode: uint
		{
			LINETERMMODE_BUTTONS = (uint)0x00000001,
			LINETERMMODE_LAMPS = (uint)0x00000002,
			LINETERMMODE_DISPLAY = (uint)0x00000004,
			LINETERMMODE_RINGER = (uint)0x00000008,
			LINETERMMODE_HOOKSWITCH = (uint)0x00000010,
			LINETERMMODE_MEDIATOLINE = (uint)0x00000020,
			LINETERMMODE_MEDIAFROMLINE = (uint)0x00000040,
			LINETERMMODE_MEDIABIDIRECT = (uint)0x00000080    
		}

    [Flags]
    public enum LineTermSharing : uint
		{
			LINETERMSHARING_PRIVATE = (uint)0x00000001,
			LINETERMSHARING_SHAREDEXCL = (uint)0x00000002,
			LINETERMSHARING_SHAREDCONF = (uint)0x00000004
		}
             
		[Flags]
		public enum LineToneMode: uint
		{
			LINETONEMODE_CUSTOM = (uint)0x00000001,
			LINETONEMODE_RINGBACK = (uint)0x00000002,
			LINETONEMODE_BUSY = (uint)0x00000004,
			LINETONEMODE_BEEP = (uint)0x00000008,
			LINETONEMODE_BILLING = (uint)0x00000010   
		}
  
		public enum LineTransferMode: uint
		{
			LINETRANSFERMODE_TRANSFER = (uint)0x00000001,
			LINETRANSFERMODE_CONFERENCE = (uint)0x00000002
		}

		[Flags]
			public enum LineTranslateOption: uint
		{
			LINETRANSLATEOPTION_CARDOVERRIDE        = (uint)0x00000001,
			LINETRANSLATEOPTION_CANCELCALLWAITING   = (uint)0x00000002,      // TAPI v1.4
			LINETRANSLATEOPTION_FORCELOCAL          = (uint)0x00000004,      // TAPI v1.4
			LINETRANSLATEOPTION_FORCELD             = (uint)0x00000008       // TAPI v1.4
		}

		[Flags]
			public enum LineTranslateResult: uint
		{
			LINETRANSLATERESULT_CANONICAL           = (uint)0x00000001,
			LINETRANSLATERESULT_INTERNATIONAL       = (uint)0x00000002,
			LINETRANSLATERESULT_LONGDISTANCE        = (uint)0x00000004,
			LINETRANSLATERESULT_LOCAL               = (uint)0x00000008,
			LINETRANSLATERESULT_INTOLLLIST          = (uint)0x00000010,
			LINETRANSLATERESULT_NOTINTOLLLIST       = (uint)0x00000020,
			LINETRANSLATERESULT_DIALBILLING         = (uint)0x00000040,
			LINETRANSLATERESULT_DIALQUIET           = (uint)0x00000080,
			LINETRANSLATERESULT_DIALDIALTONE        = (uint)0x00000100,
			LINETRANSLATERESULT_DIALPROMPT          = (uint)0x00000200,
			LINETRANSLATERESULT_VOICEDETECT         = (uint)0x00000400,      // TAPI v2.0
			LINETRANSLATERESULT_NOTRANSLATION       = (uint)0x00000800       // TAPI v3.0
		}
    
		public enum StringFormat: uint
		{
			STRINGFORMAT_ASCII = (uint)0x00000001,
			STRINGFORMAT_DBCS = (uint)0x00000002,
			STRINGFORMAT_UNICODE = (uint)0x00000003,
			STRINGFORMAT_BINARY = (uint)0x00000004   
		}
      		
		/// <summary>
		/// All Events raised by Tapi 2.2
		/// </summary>
		public enum TapiEvent
		{
			LINE_ADDRESSSTATE = 0,
			LINE_CALLINFO,
			LINE_CALLSTATE,
			LINE_CLOSE,
			LINE_DEVSPECIFIC,
			LINE_DEVSPECIFICFEATURE,
			LINE_GATHERDIGITS,
			LINE_GENERATE,
			LINE_LINEDEVSTATE,
			LINE_MONITORDIGITS,
			LINE_MONITORMEDIA,
			LINE_MONITORTONE,
			LINE_REPLY,
			LINE_REQUEST,
			PHONE_BUTTON,
			PHONE_CLOSE,
			PHONE_DEVSPECIFIC,
			PHONE_REPLY,
			PHONE_STATE,
     
			LINE_CREATE,        // Tapi v1.4                                  
			PHONE_CREATE,
    
			LINE_AGENTSPECIFIC, // Tapi v2
			LINE_AGENTSTATUS,
			LINE_APPNEWCALL,
			LINE_PROXYREQUEST,
			LINE_REMOVE,
			PHONE_REMOVE                                                                                                                                                                                               
		};
    		
		#endregion //Enumerations
	
		#region Private Members
	
		/// <summary>
		/// application friendly name to pass to LineInitializeEx
		/// </summary>
		private string m_friendlyAppName;
	
		/// <summary>
		///  handle to TAPI, set by LineInitializeEx
		/// </summary>
		private System.IntPtr m_hTapi = System.IntPtr.Zero;
	
		/// <summary>
		/// number of devices, returned by LineInitializeEx
		/// </summary>
		private uint m_numberDevices = 0;
	
		/// <summary>
		/// TAPI Version -- default to 2
		/// </summary>
		private uint m_TapiVersion = 0x00020000;
	
		/// <summary>
		/// Handle of calling application or window
		/// </summary>
		private System.IntPtr m_AppHandle = System.IntPtr.Zero;
	
		/// <summary>
		/// array of all lines found by lineInitializeEx
		/// </summary>
		private CLine[] m_CLines;		
	
		private LineCallBack m_CallBack;
	
		#endregion //  Private Members
	
		#region Asynchronous Event Handling
	
		/// <summary>
		/// this structure is for use in sinking Asynchronous events such as generated by calling lineDevSpecific
		/// the calling routine should return requestID and pinfo (which it allocates using Marshal.AllocHGlobal
		/// the process that sinks the events should maintain an ArrayList of these structs
		/// When an event occurs (e.g. LINE_REPLY) the event handler should look in the array for one with
		/// a matching requestID, and if it finds one, retrieve the required state data from pinfo
		/// NOTE: the event sink is responsible for releasing the memory using Marshal.FreeHGlobal(pinfo)
		/// </summary>
		public enum eLineReplyEventSource
		{
			lineAccept						 	=	1,
			lineAddToConference		 	=	2,
			lineAnswer						 	=	3,
			lineBlindTransfer			 	=	4,
			lineCompleteCall			 	=	5,
			lineCompleteTransfer	 	=	6,
			lineDevSpecific				 	=	8,
			lineDevSpecificFeature	=	9,
			lineDial							 	= 10,
			lineDrop								=	11,
			lineHold							 	= 12,
			lineMakeCall						=	13,
			linePark								= 14,
			linePickup							=	15,
			linePrepareAddToConference =  16,
			lineRedirect						=	 17,
			lineReleaseUserUserInfo	=	18,
			lineRemoveFromConference = 19,
			lineSecureCall					=	 20,
			lineSendUserUserInfo		=	 21,
			lineSetupTransfer				=	 22,
			lineSwapHold						=	 23,
			lineUncompleteCall			=	 24,
			lineUnhold							=	 25,
			lineUnpark							=	 26
		}
	
		public class PendingReplyEvents
		{
			public int requestID;
			public IntPtr pinfo;
			public CCall call;
			public eLineReplyEventSource source;
		}
	
		#endregion // Asynchronous Event Handling
	
		#region Constructors
	
		public CTapi(string AppName, uint Version, IntPtr AppHandle):this(AppName, Version, AppHandle,
		CTapi.LineInitializeExOptions.LINEINITIALIZEEXOPTION_USEHIDDENWINDOW)
	{}

		public CTapi(string AppName, uint Version, IntPtr AppHandle, LineInitializeExOptions opts)
		{
			// debugging
			Debug.Listeners.Add(new TextWriterTraceListener(Console.Out));
			Debug.AutoFlush = true;
			Debug.Indent();
			Debug.WriteLine("CTapi()");
    
			m_friendlyAppName = AppName;
			m_TapiVersion = Version;
			m_AppHandle = AppHandle;
			// we'll need to pass this to several API calls
			m_CallBack = new CTapi.LineCallBack(MainLineCallBack);
    
			// prepare the lineinitializeexparams
			CTapi.lineinitializeexparams initparams = new CTapi.lineinitializeexparams();
			initparams.dwTotalSize = (uint) Marshal.SizeOf(initparams);
			initparams.dwNeededSize = initparams.dwTotalSize;
			initparams.dwUsedSize = initparams.dwTotalSize;
			initparams.hEvent = System.IntPtr.Zero;
			initparams.dwOptions = (uint)opts;

			// now start up the events!
			try
			{
				///*
				CTapi.LineErrReturn ret = CTapi.lineInitializeEx(out m_hTapi, m_AppHandle, m_CallBack,
					m_friendlyAppName, out m_numberDevices, ref m_TapiVersion, ref initparams);
				//*/
				/*
				CTapi.LineInitReturn ret = CTapi.lineInitialize(out m_hTapi, m_AppHandle, m_CallBack,
					m_friendlyAppName, out m_numberDevices, ref m_TapiVersion);
				*/
				// Debugging
#if TESTING
      ShowStartupInfo(ret);
#endif

				if(ret != CTapi.LineErrReturn.LINEERR_OK)
				{
					// handle the error
				}
				ret = InitializeLines();
				//MessageBox.Show(CTapi.sLineReturn("After InitializeLines", ret));

			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message);
			}
		}
	
		~CTapi()
		{
			Dispose(false);
		}

		#endregion // Constructors

		#region Properties
	
		public LineCallBack CallBack
		{ get {return this.m_CallBack;}}

		public IntPtr hTapi
		{ get {return this.m_hTapi;}}
  
		public CLine[] Lines
		{ get {return this.m_CLines;}}
  
		#endregion // Properties
	
		#region Static Methods
	
		/// <summary>
		/// marshals a guid to the lineextensionid struct
		/// </summary>
		/// <param name="uuid"></param>
		/// <returns></returns>
		public static lineextensionid LineExtIDFromGuid(System.Guid uuid)
		{
			IntPtr buffer = Marshal.AllocHGlobal(Marshal.SizeOf(uuid));
			Marshal.StructureToPtr(uuid, buffer, false);
			//lineextensionid lineex = new CTapi.lineextensionid();
			lineextensionid lineex = (lineextensionid)Marshal.PtrToStructure(buffer, typeof(lineextensionid));
			Marshal.FreeHGlobal( buffer );
			return lineex;		  
		}

    public static LineErrReturn BlindTransfer(IntPtr hCall, string telno, out int requestID)
		{
      int retval = 0;
      requestID = 0;
			LineErrReturn val = CTapi.lineSetCallPrivilege(hCall, LineCallPrivilege.LINECALLPRIVILEGE_OWNER);
      if (val != LineErrReturn.LINEERR_OK)
        return val;
#if TESTING
			MessageBox.Show("Transfer -- set call privilege return = " + val.ToString());
#endif
      retval = CTapi.lineBlindTransfer(hCall, telno, 0);
#if TESTING
			MessageBox.Show("Blind Transfer to " + telno + " Return: " + retval.ToString());
#endif
      return HandleAsycnReturn(retval, out requestID);
		}
  
		public static LineErrReturn DeallocateCall(IntPtr hCall)
		{
			if (hCall != IntPtr.Zero)
				return lineDeallocateCall(hCall);
			else
				return LineErrReturn.LINEERR_OK;
		}

		public static LineErrReturn HangUpCall(IntPtr hCall, string userUserInfo, out int requestID)
		{
      requestID = 0;
      if (hCall != IntPtr.Zero)
      {
        int retval = lineDrop(hCall, userUserInfo, CAPIUtils.GetSize(userUserInfo));
        return HandleAsycnReturn(retval, out requestID);
      }
      else
        return LineErrReturn.LINEERR_OK;	
		}

		public static LineErrReturn SetCallPrivilege(IntPtr hCall, LineCallPrivilege priv)
		{
			return lineSetCallPrivilege(hCall, priv);
		}

		public static LineErrReturn HoldCall(IntPtr hCall, out int requestID)
		{
			int retval = LineHold(hCall);
      return HandleAsycnReturn(retval, out requestID);
		}

		public static LineErrReturn UnholdCall(IntPtr hCall, out int requestID)
		{
      requestID = 0;
      LineErrReturn ret = SetCallPrivilege(hCall, LineCallPrivilege.LINECALLPRIVILEGE_OWNER);
      if (ret != LineErrReturn.LINEERR_OK)
        return ret;
			int retval = LineUnhold(hCall);
      return HandleAsycnReturn(retval, out requestID);
    }

    public static LineErrReturn SetUserUserInfo(IntPtr hCall, string userUserInfo, out int requestID)
    {
      requestID = 0;
      LineErrReturn ret = CTapi.lineSetCallPrivilege(hCall, LineCallPrivilege.LINECALLPRIVILEGE_OWNER);
#if TESTING
			MessageBox.Show("SendUserUserInfo -- set call privilege return = " + val.ToString());
#endif
      if (ret == LineErrReturn.LINEERR_OK)
      {
        int retval = CTapi.lineSendUserUserInfo(hCall, userUserInfo, CAPIUtils.GetSize(userUserInfo));
        return HandleAsycnReturn(retval, out requestID);
      }
#if TESTING
			MessageBox.Show("SendUserUserInfo " + userUserInfo + " Return: " + retval.ToString());
#endif
      return ret;
    }
		/// <summary>
		/// returns true if the LINE_CALLINFO event contains caller id info
		/// </summary>
		/// <param name="lcis"></param>
		/// <returns></returns>
		public static bool LineCallInfoIsCallerID(LineCallInfoState lcis)
		{
			return ((lcis & LineCallInfoState.LINECALLINFOSTATE_CALLERID) != 0);
		}

		public static uint TranslateVersion(string version)
		{
			switch (version)
			{
				case "1.3":
					return 0x00010003;
				case "1.4":
					return 0x00010004;
				case "2.0":
					return 0x00020000;
				case "2.1":
					return 0x00020001;
				case "2.2":
					return 0x00020002;
				default:
					return 0x00020000;
			}
		}

		#endregion // Static Methods
	
		#region Public Methods
	
		/// <summary>
		/// finds & opens a line which meets the filter criteria
		/// </summary>
		/// <param name="Provider"></param>
		/// <param name="IsVoice"></param>
		/// <param name="priv"></param>
		/// <returns></returns>
		public LineErrReturn GetLineByFilter(string Provider, bool IsVoice, CTapi.LineCallPrivilege priv,
			out CLine line)
		{
			line = null;
			bool lineOK = false;
			for (int i = 0; i < this.Lines.Length; ++i)
			{
				line = this.Lines[i];
				lineOK = false;
				// first make sure it meets the basics!
				if ((line.IsCompatible == true) && (line.LineValid == true))
				{
					if(!String.IsNullOrEmpty(Provider))
					{
						if(line.ProviderInfo.IndexOf(Provider) >= 0)
							lineOK = true;
					}
					else
						lineOK = true;
	        
					// if we get here and lineOK == true, either LineName was null, or the name met the filter criteria		        
					if (lineOK == true)
					{
						if (IsVoice == true)
							lineOK = line.IsVoiceLine();
					}
				}
	    
				if(lineOK == true)
				{
					// open the line and return it
					CTapi.LineErrReturn ret;
					// first get the address caps for the line
					ret = line.FillAddressCaps();
					if(ret != CTapi.LineErrReturn.LINEERR_OK)
						return ret;
#if TESTING
				for (int iaddr = 0; iaddr < line.NumAddresses; ++iaddr)
				{
						
						MessageBox.Show("Line = " + line.DeviceID + " Address = " + iaddr + " Address ExtVer = "
							+ line.Addresses[iaddr].ExtVersion);
						
				}
				MessageBox.Show(CTapi.sLineReturn("After FillAddressCaps", ret));
#endif
					ret = line.LineOpen(line.DeviceID, priv);
					return ret;
				}  
			}
			return LineErrReturn.LINEERR_INVALDEVICECLASS;
		}
		/// <summary>
		/// 
		/// </summary>
		/// <param name="Provider"></param>
		/// <param name="IsVoice"></param>
		/// <param name="priv"></param>
		/// <returns></returns>
		public LineCollection GetVoiceLinesByFilter(string Provider, LineCallPrivilege priv)
		{
			LineCollection lineCollection = new LineCollection();
			CLine line;
			bool lineOK = false;
			for (int i = 0; i < this.Lines.Length; ++i)
			{
				line = this.Lines[i];
				lineOK = false;
				// first make sure it meets the basics!
				if ((line.IsCompatible == true) && (line.LineValid == true))
				{
          if (!String.IsNullOrEmpty(Provider))
					{
						if(line.ProviderInfo.IndexOf(Provider) >= 0)
							lineOK = true;
					}
					else
						lineOK = true;
	        
					// if we get here and lineOK == true, either LineName was null, or the name met the filter criteria		        
					if (lineOK == true)
					{
						lineOK = line.IsVoiceLine();
					}
				}
	    
				if(lineOK == true)
				{
					// open the line and return it
					LineErrReturn ret;
					// first get the address caps for the line
					ret = line.FillAddressCaps();
          if (ret == CTapi.LineErrReturn.LINEERR_OK)
          {
            ret = line.LineOpen(line.DeviceID, priv);
            if (ret == CTapi.LineErrReturn.LINEERR_OK)
              lineCollection.Add(line);
          }  
				}  
			}
			return lineCollection;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public LineCollection GetVoiceLines()
		{
			LineCollection lineCollection = new LineCollection();
			CLine line;
			for (int i = this.Lines.GetLowerBound(0); i < this.Lines.GetUpperBound(0) +1; i++)
			{
				line = this.Lines[i];
				if(line.IsVoiceLine())
					lineCollection.Add(line);
		
			}
			return lineCollection;
		}

		//	the DialNumber has to be adjusted to the telephony settings (dial prefix ..)
		internal LineErrReturn GetTranslateAddress(CLine line, string DialNumber, ref linetranslateoutput TranslateOutput, ref byte[] totalbuffer )
		{
			int buffsize			= 1;
			int InitSize			= Marshal.SizeOf(TranslateOutput);
			int TotalSize			= InitSize;
			int NeededSize			= InitSize;
			bool needMoreMem		= true;
			LineErrReturn ret		= LineErrReturn.LINEERR_STRUCTURETOOSMALL;
			IntPtr pTransOutput;

			while (needMoreMem == true)
			{
				if((NeededSize - TotalSize) > 0)
					buffsize = (NeededSize - InitSize + 1);
				else
					buffsize = 1;

				TranslateOutput.dwTotalSize = (uint)(buffsize + InitSize - 1);
				pTransOutput = Marshal.AllocHGlobal(NeededSize);
				Marshal.StructureToPtr(TranslateOutput, pTransOutput, true);
				ret = lineTranslateAddress(this.m_hTapi, line.DeviceID, this.m_TapiVersion, DialNumber, 0, 0, pTransOutput);
				Marshal.PtrToStructure(pTransOutput, TranslateOutput);
				TotalSize	= (int)TranslateOutput.dwTotalSize;
				NeededSize	= (int)TranslateOutput.dwNeededSize;  
				needMoreMem = ((ret == LineErrReturn.LINEERR_OK) &&(NeededSize > TotalSize)) || (ret == LineErrReturn.LINEERR_STRUCTURETOOSMALL);
				if ( needMoreMem == false )
				{
					totalbuffer = new byte[TotalSize];
					Marshal.Copy(pTransOutput,totalbuffer,0,TotalSize);
				}
				Marshal.FreeHGlobal(pTransOutput);

			}
			return ret;
		}

		public LineErrReturn MakeCall (CLine line, string DialNumber, out IntPtr hCall, out int requestID)
		{
			int retval = lineMakeCall(line.hLine, out hCall, DialNumber, 0, IntPtr.Zero);
      return HandleAsycnReturn(retval, out requestID);
		}

		//	dials the number and returns whether this was succesful
		// note: this is an asynchronous function, so must correlate the output requestID
		// with the status return of the LINE_REPLY event
    public LineErrReturn DialNumber(CLine Line, string DialNumber, out IntPtr hCall, out int requestID)
		{
			LineErrReturn	result = LineErrReturn.LINEERR_OK;
			linetranslateoutput	l_TranslateOutput	= new linetranslateoutput();
			linecallparams l_callparams = new linecallparams();
			uint l_SizeOfCallParams;
			string l_DisplayablePhoneNum;
			string l_DialablePhoneNum;
			int l_TransSize;
			byte[] l_buffer	= null;

      requestID = 0;
			hCall = IntPtr.Zero;
			System.Text.Decoder dec	= null;
			StringFormat sf	= Line.StringFormat;

			result = GetTranslateAddress(Line, DialNumber, ref l_TranslateOutput, ref l_buffer );
			if ( result != LineErrReturn.LINEERR_OK )
			{
				return result;
			}

			dec	= GetDecoder(sf);
			l_TransSize	= Marshal.SizeOf(l_TranslateOutput);

			l_DisplayablePhoneNum = CAPIUtils.StringFromByteArray(l_buffer, 
				(int)l_TranslateOutput.dwDisplayableStringOffset,
				(int)l_TranslateOutput.dwDisplayableStringSize, dec );
			l_DialablePhoneNum = CAPIUtils.StringFromByteArray(l_buffer, 
				(int)l_TranslateOutput.dwDialableStringOffset,
				(int)l_TranslateOutput.dwDialableStringSize, dec );

			l_SizeOfCallParams = (uint)Marshal.SizeOf(l_callparams);

			// Set the Call Parameters.
			l_callparams.dwTotalSize = l_SizeOfCallParams+l_TranslateOutput.dwDisplayableStringSize;
			l_callparams.dwBearerMode = (uint)LineBearerMode.LINEBEARERMODE_VOICE;
			l_callparams.dwMediaMode = (uint)LineMediaMode.LINEMEDIAMODE_INTERACTIVEVOICE; 
			l_callparams.dwCallParamFlags = (uint)LineCallParamFlags.LINECALLPARAMFLAGS_IDLE;
			l_callparams.dwAddressMode = (uint)LineAddressMode.LINEADDRESSMODE_ADDRESSID;
			l_callparams.dwAddressID = Line.Addresses[0].AddressID;
			l_callparams.dwDisplayableAddressSize	= l_TranslateOutput.dwDisplayableStringSize;
			l_callparams.dwDisplayableAddressOffset = l_SizeOfCallParams;
				
			//	copy the Call Parameters structure and the DisplayablePhoneNum to the IntPtr
			//int rawsize = (int)l_SizeOfCallParams + (int)l_TranslateOutput.dwDisplayableStringSize;
			int rawsize = (int)l_SizeOfCallParams + (int)l_TranslateOutput.dwDialableStringSize;
			IntPtr l_ParamBuffer = Marshal.AllocHGlobal( rawsize );
			//IntPtr l_ParamBuffer = DebugMemHandler.Instance.AllocHGlobal(rawsize, 
			//	"CTapi::DialNumber,l_ParamBuffer"); 
			IntPtr l_PhoneBuffer = Marshal.AllocHGlobal(
				(int)l_TranslateOutput.dwDialableStringSize);
			//IntPtr l_PhoneBuffer = DebugMemHandler.Instance.AllocHGlobal(
			//	(int)l_TranslateOutput.dwDialableStringSize /*dwDisplayableStringSize*/, 
			//	"CTapi::DialNumber,l_PhoneBuffer"); 
			Marshal.StructureToPtr( l_callparams, l_ParamBuffer, false );
			Marshal.Copy( l_buffer, (int)l_TranslateOutput.dwDialableStringOffset, 
				l_PhoneBuffer, (int)l_TranslateOutput.dwDialableStringSize );
			Marshal.WriteIntPtr( l_ParamBuffer, (int)l_SizeOfCallParams, l_PhoneBuffer );
			Marshal.FreeHGlobal( l_PhoneBuffer );

			int retval = lineMakeCall( Line.hLine, out hCall, l_DialablePhoneNum, 0, l_ParamBuffer);

			Marshal.FreeHGlobal( l_ParamBuffer );
      return HandleAsycnReturn(retval, out requestID);

		}

    public LineErrReturn GenerateDTMFDigits(IntPtr hCall, string digits)
    {
      return lineGenerateDigits(hCall, LineDigitMode.LINEDIGITMODE_DTMF, digits, 0);
    }

		/// <summary>
		/// helper method to get a decoder to use when parsing byte data into properties
		/// </summary>
		/// <param name="sf"></param>
		/// <returns></returns>
		public static System.Text.Decoder GetDecoder(CTapi.StringFormat sf)
		{
			System.Text.Encoding enc;
			System.Text.Decoder dec = null;
		
			// get the right decoder
			switch (sf)
			{
				case CTapi.StringFormat.STRINGFORMAT_ASCII:
					enc = new System.Text.ASCIIEncoding();
					dec = enc.GetDecoder();
					break;
		  
				case CTapi.StringFormat.STRINGFORMAT_BINARY:
					// no encoding or decoding; just an array of bytes
					// this is a special case that will have to be handled when needed
					break;
		  
				case CTapi.StringFormat.STRINGFORMAT_DBCS:
					enc = Encoding.GetEncoding(1252);
					dec = enc.GetDecoder();
					break;
		  
				case CTapi.StringFormat.STRINGFORMAT_UNICODE:
					enc = new System.Text.UnicodeEncoding();
					dec = enc.GetDecoder();
					break;
			}
			return dec;
		}		
			
		/// <summary>
		/// returns the string associated with the TapiEvent enum
		/// </summary>
		/// <param name="ev"></param>
		/// <returns></returns>
		public static string sTapiEvent (TapiEvent ev)
		{ 
			return Enum.Format(typeof(TapiEvent), ev, "G");
		}
  
		/// <summary>
		/// returns friendly string value for CallState enum
		/// </summary>
		/// <param name="lcs"></param>
		/// <returns></returns>
		public static string sCallState (CTapi.LineCallState lcs)
		{ 
			return Enum.Format(typeof(CTapi.LineCallState), lcs, "G");
		}
  
		/// <summary>
		/// returns the friendly name of the lineerr return
		/// </summary>
		/// <param name="Caller"></param>
		/// <param name="ret"></param>
		/// <returns></returns>
		public static string sLineReturn(string Caller, LineErrReturn ret)
		{
			return "Caller = " + Caller + " return = " + Enum.Format(typeof(LineErrReturn), ret, "G");
		}
  
		/// <summary>
		/// returns a friendly string from the StringFormat enum
		/// </summary>
		/// <param name="f"></param>
		/// <returns></returns>
		public static string sStringFormat(StringFormat f)
		{
			return Enum.Format(typeof(StringFormat), f, "G");
		}
  
		/// <summary>
		/// shows return values after call from LineInitializeEx -- useful for debugging
		/// </summary>
		/// <param name="ret"></param>
		public void ShowStartupInfo(CTapi.LineErrReturn ret)
		{
			StringBuilder sb = new StringBuilder(100);
			sb.Append("Return from lineInitializeEx: " + ret.ToString("X") + Environment.NewLine);
			sb.Append("Tapi Handle: " + m_hTapi.ToString() + Environment.NewLine);
			sb.Append("Number of lines = " + m_numberDevices + Environment.NewLine);
			MessageBox.Show (sb.ToString());

		}
      
		/// <summary>
		/// caller should clean up by calling this in the Dispose method
		/// </summary>
		/// <returns></returns>
		public CTapi.LineErrReturn ShutDown()
		{
			this.CloseLines();
			if (this.m_hTapi != IntPtr.Zero)
				return lineShutdown(this.m_hTapi);
			else
				return CTapi.LineErrReturn.LINEERR_OK;     
		}
  
		#endregion // Public Methods
  
		#region Private Methods

    /// <summary>
    /// Helper method to handle returns from Tapi methods that are completed asynchronously
    /// </summary>
    /// <param name="retval"></param>
    /// <param name="requestID"></param>
    /// <returns></returns>
    private static CTapi.LineErrReturn HandleAsycnReturn(int retval, out int requestID)
    {
      requestID = 0;
      if (retval > 0)
      {
        requestID = retval;
        return LineErrReturn.LINEERR_OK;
      }
      else
      {
        return (LineErrReturn)(-retval);
      }
    }
  
		/// <summary>
		/// fills in information for all lines from lineNegotiateAPIVersion
		/// TODO: Consider moving this to CLine
		/// </summary>
		/// <returns></returns>
		private CTapi.LineErrReturn InitializeLines()
		{
			CTapi.LineErrReturn retDevCaps;
			CTapi.lineextensionid LineExtensionID;
			uint TapiVer;
			CLine line;
			this.m_CLines = new CLine[this.m_numberDevices];
	  
			for (int i = 0; i < this.m_numberDevices; i++)
			{
				line = new CLine(this, (uint)i);
				CTapi.LineErrReturn ret = CTapi.lineNegotiateAPIVersion(this.m_hTapi, (uint)i, CTapi.TAPILOWVERSION,
					CTapi.TAPIHIVERSION, out TapiVer, out LineExtensionID);
				if(ret == CTapi.LineErrReturn.LINEERR_OK)
				{
					line.m_TapiNegotiatedVer = TapiVer;
					line.m_LineExtensionID = LineExtensionID;
					line.m_IsCompatible = true;
	      
					CTapi.LineErrReturn extret = CTapi.lineNegotiateExtVersion(this.m_hTapi, (uint) i,
						line.TapiNegotiatedVer, CTapi.TSPLOWEXTVERSION, CTapi.TSPHIEXTVERSION, out line.m_ExtVersion);
					// debugging
#if TESTING
				MessageBox.Show("Line = " + i + Environment.NewLine
					+ "TSP LO Ver = " + CTapi.TSPLOWEXTVERSION.ToString()+ Environment.NewLine
					+ "Negotiated Ver = " + line.TapiNegotiatedVer.ToString()+ Environment.NewLine
					+ "Ext Ver = " + line.m_ExtVersion.ToString());
#endif
					if (extret != CTapi.LineErrReturn.LINEERR_OK)
					{
						line.m_ExtVersion = 0; 
					}
					retDevCaps = line.GetDeviceCaps();
					if (retDevCaps != CTapi.LineErrReturn.LINEERR_OK)
					{
						// handle error!
						line.LineValid = false;
#if TESTING
	        MessageBox.Show("Return from CLine.GetDeviceCaps for line " + i + " = " + retDevCaps.ToString());
#endif
					}
					else
					{
#if TESTING
	        MessageBox.Show("Return from CLine.GetDeviceCaps for line " + i + " = OK" );
	        MessageBox.Show (line.GetLineInfoString());
#endif
					}
				}
				else
				{
					line.m_IsCompatible = false;
				}
				m_CLines[i] = line;
			}
			return CTapi.LineErrReturn.LINEERR_OK;
		}
	   
		private void CloseLines()
		{
			for (int i = 0; i < this.m_CLines.Length; ++i)
			{
				this.m_CLines[i].LineClose();
			}
		}

	/*
		private LineErrReturn DeallocateCall(IntPtr hCall)
		{
			if (hCall != IntPtr.Zero)
				return lineDeallocateCall(hCall);
			else
				return LineErrReturn.LINEERR_OK;
		}
  */
		#endregion // Private Methods
	
		#region Internal Methods
	


		#endregion // Internal Methods
	
		#region IDisposable

		public void Dispose()
		{
			GC.SuppressFinalize(this);
			Dispose(true);
		}
	
		/// <summary>
		/// just in case the client prefers to call Close instead of Dispose
		/// </summary>
		public void Close()
		{
			Dispose();
		}
	
		private void Dispose(bool disposing)
		{
			lock(this)
			{
				if(disposing)
				{
					this.ShutDown();
				}
			}
		}

		#endregion // IDisposable

		#region Event Handling
	
		public void MainLineCallBack(System.UInt32 dwDevice, System.UInt32 dwMessage, System.UInt32 dwInstance,
			System.UInt32 dwParam1, System.UInt32 dwParam2, System.UInt32 dwParam3)
		{
			// Debugging
#if TESTING
	  //MessageBox.Show(CTapi.sTapiEvent((CTapi.TapiEvent)dwMessage));
	  Debug.WriteLine(CTapi.sTapiEvent((CTapi.TapiEvent)dwMessage));
#endif
			IntPtr hCall = IntPtr.Zero;

			switch ((CTapi.TapiEvent)dwMessage)
			{
				case CTapi.TapiEvent.LINE_ADDRESSSTATE:
					break;
				case CTapi.TapiEvent.LINE_CALLINFO:
					LineCallInfoState lcis = (LineCallInfoState)dwParam1;
					hCall = (IntPtr)dwDevice;
					OnCallInfoEvent(new LineCallInfoEventArgs(hCall, lcis));
					break;
				case CTapi.TapiEvent.LINE_CALLSTATE:
				{
					LineCallState lcs = (LineCallState) dwParam1;
					LineErrReturn ret;
					hCall = (IntPtr)dwDevice;
					OnCallStateEvent(new CallStateEventArgs(hCall, new CCall(this, hCall), lcs, dwParam2));
					// if it's IDLE, deallocate the call after all event handling
					if (lcs == LineCallState.LINECALLSTATE_IDLE)
					{
						ret = DeallocateCall(hCall);
						// what do we want to do if ret is not OK?
					}
				}
					break;                          
				case CTapi.TapiEvent.LINE_CLOSE:
					break;                          
				case CTapi.TapiEvent.LINE_DEVSPECIFIC:
					break;                          
				case CTapi.TapiEvent.LINE_DEVSPECIFICFEATURE:
					break;                          
				case CTapi.TapiEvent.LINE_GATHERDIGITS:
          OnGatherDigits(new LineGatherDigitsEventArgs((IntPtr)dwDevice, (LineGatherTerm)dwParam1));

					break;                          
				case CTapi.TapiEvent.LINE_GENERATE:
					break;                          
				case CTapi.TapiEvent.LINE_LINEDEVSTATE:
					break;                          
				case CTapi.TapiEvent.LINE_MONITORDIGITS:
					break;                          
				case CTapi.TapiEvent.LINE_MONITORMEDIA:
					break;                          
				case CTapi.TapiEvent.LINE_MONITORTONE:
					break;                          
				case CTapi.TapiEvent.LINE_REPLY:
					OnLineReply(new LineReplyEventArgs(dwParam1, dwParam2));
					break;                          
				case CTapi.TapiEvent.LINE_REQUEST:
					break;                          
				case CTapi.TapiEvent.PHONE_BUTTON:
					break;                          
				case CTapi.TapiEvent.PHONE_CLOSE:
					break;                          
				case CTapi.TapiEvent.PHONE_DEVSPECIFIC:
					break;                          
				case CTapi.TapiEvent.PHONE_REPLY:
					break;                          
				case CTapi.TapiEvent.PHONE_STATE:
					break;                          
				case CTapi.TapiEvent.LINE_CREATE:
					break;                          
				case CTapi.TapiEvent.PHONE_CREATE:
					break;                          
				case CTapi.TapiEvent.LINE_AGENTSPECIFIC:
					break;                          
				case CTapi.TapiEvent.LINE_AGENTSTATUS:
					break;                          
				case CTapi.TapiEvent.LINE_APPNEWCALL:
				{
					hCall = (IntPtr)dwParam2;
					IntPtr hLine = (IntPtr)dwDevice;
					uint LineDeviceID = dwParam1;
					OnAppNewCall(new AppNewCallEventArgs(this.m_CLines[(int)LineDeviceID], new CCall(this, hCall)));		      
				}
					break;                          
				case CTapi.TapiEvent.LINE_PROXYREQUEST:
					break;                          
				case CTapi.TapiEvent.LINE_REMOVE:
					break;                          
				case CTapi.TapiEvent.PHONE_REMOVE:
					break;                          
			} // end switch
	  
		}
	
		public virtual void OnAppNewCall(AppNewCallEventArgs e)
		{
			if (AppNewCall != null)
				AppNewCall(this, e);
		}
	
		public virtual void OnCallStateEvent(CallStateEventArgs e)
		{
			if(CallStateEvent != null)
				CallStateEvent(this, e);
		}
	
		public virtual void OnLineReply(LineReplyEventArgs e)
		{
			if(LineReplyEvent != null)
				LineReplyEvent(this, e);
		}
	
		public virtual void OnCallInfoEvent(LineCallInfoEventArgs e)
		{
			if(LineCallInfoEvent != null)
				LineCallInfoEvent(this, e);
		}

    public virtual void OnGatherDigits(LineGatherDigitsEventArgs e)
    {
      if (LineGatherDigits != null)
        LineGatherDigits(this, e);
    }

		#endregion // Event Handling		
	
	} // class CTapi
} // namespace AgileSoftwareInc.Tapi2Lib
