// Shoreline Teleworks Network Call Control TSP Extension

#include <winsock.h>

#define NCC_EXT_VERSION		((DWORD) 0x00010001)

#define NCC_LINE_EXTENSION_ID \
	{ /* D9E8E4C0-11D1-11D8-6000-478E19853797 */ \
		0xD9E8E4C0, \
		0x11D811D1, \
		0x8E470060, \
		0x97378519  \
	}

#define NCC_PHONE_EXTENSION_ID \
	{ /* D9E8E4C1-11D1-11D8-6000-478E19853797 */ \
		0xD9E8E4C1, \
		0x11D811D1, \
		0x8E470060, \
		0x97378519  \
	};

#ifndef GUID_DEFINED
#define GUID_DEFINED
typedef struct _GUID
{
    unsigned long  Data1;
    unsigned short Data2;
    unsigned short Data3;
    unsigned char  Data4[8];
} GUID;
#endif /* GUID_DEFINED */

////////////////////////////////////////////////////////
// TAPI Line Device Extensions

// @@ Begin // This section is needed by java

// Device Specific Line Functions
#define STLINE_SENDTOMAILBOX				0x00000001
#define STLINE_GETCALLTUPLES				0x00000002
#define STLINE_SETCALLTUPLES				0x00000003
#define STLINE_DELETECALLTUPLES				0x00000004
#define STLINE_GETLINETUPLES				0x00000005
#define STLINE_GETTRUNKINFO					0x00000006
#define STLINE_GETPARTIESLINEINFO			0x00000007
#define STLINE_GETSERVERTIMEBASE            0x00000008
#define STLINE_XCHGMEDIAPORTS	            0x00000009
#define STLINE_SETMONITOREDDNS				0x0000000A
#define STLINE_SETSHORELOCATIONINFO			0x0000000B
#define STLINE_SETMESSAGEWAITING			0x0000000C
#define STLINE_SETLINETUPLES				0x0000000D
#define STLINE_DELETELINETUPLES				0x0000000E
#define STLINE_STOPIGNOREDTMF				0x0000000F
#define STLINE_XFERTOSITE		            0x00000010
#define STLINE_MEDIACTL2		            0x00000011
#define STLINE_GETDNISINFO		            0x00000012
#define STLINE_SETPOLYCOMSTATE	            0x00000013
#define STLINE_GETCALLGUID					0x00000014
#define STLINE_SETLOCALCALLTUPLES			0x00000015
#define STLINE_GETCALLTYPE					0x00000016
#define	STLINE_BLINDXFER					0x00000017

// Device Specific Line Events
#define STLINE_CALLTUPLE					0x00000001
#define STLINE_LINETUPLE					0x00000002
#define STLINE_SERVERTIMEBASECHANGE			0x00000003
	// dwParam1	= (DWORD) STLINE_SERVERTIMEBASECHANGE
	// dwParam2	= (DWORD) FILETIME.dwHighDateTime		current UTC time on server
	// dwParam3	= (DWORD) FILETIME.dwLowDateTime

// Call Tuple (1) change mask
#define STLINE_CALLTUPLE_UNKNOWN			(1 << 0)
#define STLINE_CALLTUPLE_ROUTESLIP			(1 << 1)
#define STLINE_CALLTUPLE_CALLNOTE			(1 << 2)
#define STLINE_CALLTUPLE_CALLQUEUING		(1 << 3)
#define STLINE_CALLTUPLE_CALLPOSITION		(1 << 4)
#define STLINE_CALLTUPLE_ACC_CALLCODE		(1 << 5)
#define STLINE_CALLTUPLE_ACC_CODEREQUEST	(1 << 6)

// Device Specific Line Errors
#define STLINEERR_INVALPHONESTATE			0xC0000001
#define STLINEERR_OPERATIONUNAVAIL			0xC0000002
#define STLINEERR_INVALCALLSTATEPENDING		0xC0000003
#define STLINEERR_INVALIDTUPLEARRAY			0xC0000004
#define STLINEERR_INVALIDTUPLE				0xC0000005
#define STLINEERR_NOCONFERENCE_VM			0xC0000006

#define STLINE_WSTRING_SHORELINELOCATIONNAME (L"Shoreline System - Do Not Modify")

// @@ End // This section is needed by java

// Device Specific Line Function Structure
typedef struct STLINEFUNC				STLINEFUNC;
typedef struct STLINEFUNC*				LPSTLINEFUNC;

typedef struct STLINESENDTOMAILBOX		STLINESENDTOMAILBOX;
typedef struct STLINESENDTOMAILBOX*		LPSTLINESENDTOMAILBOX;

typedef struct STLINEBLINDXFER			STLINEBLINDXFER;
typedef struct STLINEBLINDXFER*			LPSTLINEBLINDXFER;

typedef struct STLINESETMONITOREDDNS	STLINESETMONITOREDDNS;
typedef struct STLINESETMONITOREDDNS*   LPSTLINESETMONITOREDDNS;

typedef struct STLINEXFERTOSITE			STLINEXFERTOSITE;
typedef struct STLINEXFERTOSITE*		LPSTLINEXFERTOSITE;

typedef struct STMEDIACTL2				STMEDIACTL2;
typedef struct STMEDIACTL2*				LPSTMEDIACTL2;

typedef struct STDNISINFO				STDNISINFO;
typedef struct STDNISINFO*				LPSTDNISINFO;

typedef struct STLINESETSHORELOCATIONINFO	STLINESETSHORELOCATIONINFO;
typedef struct STLINESETSHORELOCATIONINFO*	LPSTLINESETSHORELOCATIONINFO;

typedef struct STLINESETMESSAGEWAITING	STLINESETMESSAGEWAITING;
typedef struct STLINESETMESSAGEWAITING*	LPSTLINESETMESSAGEWAITING;

typedef struct STLINESETPOLYCOMSTATE	STLINESETPOLYCOMSTATE;
typedef struct STLINESETPOLYCOMSTATE*	LPSTLINESETPOLYCOMSTATE;

typedef struct STLINEGETCALLGUID		STLINEGETCALLGUID;
typedef struct STLINEGETCALLGUID*		LPSTLINEGETCALLGUID;

typedef struct STLINEGETCALLTYPE		STLINEGETCALLTYPE;
typedef struct STLINEGETCALLTYPE*		LPSTLINEGETCALLTYPE;	

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINEFUNC
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINESENDTOMAILBOX
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID
	DWORD					dwDestAddressSize;
	DWORD					dwDestAddressOffset;
	DWORD					dwCountryCode;
};

enum EXferFlag
{
	eXferFlag_Null						= 0x000,
	eXferFlag_DontUpdateMailBox			= 0x001,
	eXferFlag_DontTryBackup				= 0x002,
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINEBLINDXFER
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID
	DWORD					dwDestAddressSize;
	DWORD					dwDestAddressOffset;
	EXferFlag				eBlindXferFlag;
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINEGETSERVERTIMEBASE
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID
	SYSTEMTIME				systemTime;
	TIME_ZONE_INFORMATION	timeZoneInfo;
	DWORD                   dwTimeZoneIDType;
	DWORD					dwTickCount;
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINEXCHGMEDIAPORTS
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID (STLINE_XCHGMEDIAPORTS)
	IN_ADDR					vmIPAddr;			// VoiceMail Server RTP Source IP Address
	WORD					vmSrcPort;			// VoiceMail Server RTP Source Port Number
	WORD					vmDstPort;			// VoiceMail Server RTP Destination Port Number
	IN_ADDR					ipbxIPAddr;			// IPBX RTP Source IP Address
	WORD					ipbxSrcPort;		// IPBX RTP Source Port Number
	WORD					ipbxDstPort;		// IPBX RTP Destination Port Number
};


struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINESETMONITOREDDNS
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID (STLINE_XCHGMEDIAPORTS)
	DWORD					dwDNsSize;			// size of the buffer
	DWORD					dwDNsOffset;		// offset of buffer that contains comma delimitated DNs
};


struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINEXFERTOSITE
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID (STLINE_XFERTOSITE)
	DWORD					dwDestAddressSize;
	DWORD					dwDestAddressOffset;
	DWORD					dwSiteID;			// Destination Site ID
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STMEDIACTL2
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID (STLINE_MEDIACTL2)
	DWORD					dwMediaFunction;	// Enumerated Media Function
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STDNISINFO
{
	LINEEXTENSIONID		extensionID;		// ShoreTel TAPI Extension GUID
	DWORD			dwFunction;			// Device Specific Function ID (STLINE_GETDNISINFO)
	WCHAR			wcDnisDigits[16];		// DNIS digits received from Central Office
	WCHAR			wcDnisDescription[16];		// DNIS description matched in configuration database
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINESETSHORELOCATIONINFO
{
	LINEEXTENSIONID		extensionID;		// ShoreTel TAPI Extension GUID
	DWORD			dwFunction;		// Device Specific Function ID
	DWORD			dwLocationID;		// location ID of the Shoreline location
							// - negative value means must create. 
	DWORD			dwCountryCode;
	WCHAR			wcAreaCode[16];
	WCHAR			wcLocalAccessCode[16];
	WCHAR			wcLongDistanceAccessCode[16];
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINESETMESSAGEWAITING
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID
	WCHAR					wcExtension[16];	// Address of targetted extension
	BOOL					bQueryOnly;			// Do not change only query
	BOOL					bMWI;				// MWI status on/off.  Inbound -- new value, Outbound -- old value.
};

#ifndef EPolycomState_DEFINED
#define EPolycomState_DEFINED
enum EPolycomState
{
	ePolycomState_Null			= 0x000,
	ePolycomState_HookSwitch	= 0x001,
	ePolycomState_Display		= 0x002,
	ePolycomState_SoftKey0		= 0x004,
	ePolycomState_SoftKey1		= 0x008,
	ePolycomState_SoftKey2		= 0x010,
	ePolycomState_SoftKey3		= 0x020,
	ePolycomState_CallApp1		= 0x040,
	ePolycomState_CallApp2		= 0x080,
	ePolycomState_CallApp3		= 0x100,
	ePolycomState_Alert_Lamp	= 0x200,
	ePolycomState_Dnd_Icon		= 0x400,
};
#endif

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINESETPOLYCOMSTATE
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID
	DWORD					dwStateMask;		// OR'ed EPolycomState mask bits indicate which fields are used
	DWORD					dwHookSwitchDevice;	// Active Hook Switch Device
	DWORD					dwHookSwitchMode;	// Hook Switch Mode (ONHOOK implies all hook switches are on hook).
	WCHAR					wcDisplay[3*19];	// Main Display Text (3 Rows x 19 Columns)
	WCHAR					wcSoftKey[4][8];	// SoftKey 0-3 Information (Text & Function)
	DWORD					dwButtonFunction[4];// SoftKey 0-3 Information (Text & Function)
	DWORD					dwLampMode[5];		// Lamp Modes
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINEGETCALLGUID
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID
	DWORD					dwCallID;			// 32-bit Call ID DWORD
	GUID					guidCallID;			// 128-bit Call ID GUID
};

// Call Types
enum
{
	eCALLTYPE_NULL				= 0x000,
	eCALLTYPE_HUNT				= 0x001,
	eCALLTYPE_GROUP				= 0x002,
	eCALLTYPE_OTHER				= 0x003
};

struct __declspec(uuid("{D9E8E4C0-11D1-11D8-6000-478E19853797}")) STLINEGETCALLTYPE
{
	LINEEXTENSIONID			extensionID;		// ShoreTel TAPI Extension GUID
	DWORD					dwFunction;			// Device Specific Function ID
	DWORD					dwCallType;			// Call Type
};

// Media Control2 functions:
enum
{
	eMEDIACTL2_NULL				= 0,
	eMEDIACTL2_PLAY_START		= 1,
	eMEDIACTL2_PLAY_STOP		= 2,
	eMEDIACTL2_RECORD_START		= 3,
	eMEDIACTL2_RECORD_STOP		= 4,
};

// @@ Begin // This section is needed by java

////////////////////////////////////////////////////////
// TAPI Phone Device Extensions

// Device Specific Phone Functions
#define STPHONE_SETSTATUSMESSAGES			0x00000001

// Device Specific Phone Messages
#define STPHONE_STATE						0x00000001
	// dwParam1	= (DWORD) STPHONE_STATE;
	// dwParam2	= (DWORD) STPHONESTATE_ <value>;
	// dwParam3	= (DWORD) 0;

#define STPHONE_PHONEBUTTONINFO				0x00000002
	// dwParam1	= (DWORD) STPHONE_PHONEBUTTONINFO;
	// dwParam2	= (DWORD) idButtonLamp;
	// dwParam3	= (DWORD) 0;

#define STPHONE_LINETUPLE					0x00000003
	// dwParam1	= (DWORD) STPHONE_LINETUPLE;
	// dwParam2	= (DWORD) 0;
	// dwParam3	= (DWORD) 0;

#define STPHONE_TRUNK_DIAL					0x00000004
	// dwParam1	= (DWORD) STPHONE_TRUNK_DIAL;
	// dwParam2	= (DWORD) 0;
	// dwParam3	= (DWORD) 0;

// Device Specific Phone Errors
#define STPHONEERR_INVALPHONESTATE			0xD0000001
#define STPHONEERR_OPERATIONUNAVAIL			0xD0000002
#define STPHONEERR_DEVICEOFFLINE			0xD0000003
// STPHONEERR_DEVICEOFFLINE -- This indicates that the physical device is currently offline, however the device state change that was posted has been cached and will be sent to the device when it becomes available.


// Device Specific Phone States	(EXTENSIONS)
#define STPHONESTATE_OTHER					0x00000001
#define STPHONESTATE_ON_HOOK				0x00000002
#define STPHONESTATE_RINGING				0x00000004
#define STPHONESTATE_OFF_HOOK				0x00000008
#define STPHONESTATE_PROCEEDING				0x00000010
#define STPHONESTATE_INTERNAL_DIALTONE		0x00000020
#define STPHONESTATE_EXTERNAL_DIALTONE		0x00000040
#define STPHONESTATE_DIALING				0x00000080
#define STPHONESTATE_ESTABLISHED			0x00000100
#define STPHONESTATE_REORDER				0x00000200
#define STPHONESTATE_INTERCEPT				0x00000400
#define STPHONESTATE_BUSY					0x00000800
#define STPHONESTATE_RINGBACK				0x00001000

// Device Specific Phone States	(TRUNKS)
#define STTRUNKSTATE_OTHER					0x80000000
#define STTRUNKSTATE_IDLE					0x40000000
#define STTRUNKSTATE_RINGING				0x20000000
#define STTRUNKSTATE_ANSWERING				0x10000000
#define STTRUNKSTATE_SEIZING				0x08000000
#define STTRUNKSTATE_DIALING				0x04000000
#define STTRUNKSTATE_ESTABLISHED			0x02000000
#define STTRUNKSTATE_RELEASING				0x01000000
#define STTRUNKSTATE_OOSERVICE				0x00800000
#define STTRUNKSTATE_REORDER				0x00400000
#define STTRUNKSTATE_INTERCEPT				0x00200000
#define STTRUNKSTATE_BUSY					0x00100000
#define STTRUNKSTATE_RINGBACK				0x00080000

// @@ End // This section is needed by java

// Device Specific Phone Function Structure
typedef struct STPHONEFUNC		STPHONEFUNC;
typedef struct STPHONEFUNC*		LPSTPHONEFUNC;
struct __declspec(uuid("{D9E8E4C1-11D8-11D1-8E47-006097378519}")) STPHONEFUNC
{
	PHONEEXTENSIONID	extensionID;
	DWORD				dwFunction;
	DWORD				dwParam0;
	DWORD				dwParam1;
	DWORD				dwParam2;
};

// PHONECAPS DevSpecific
typedef struct STPHONECAPS		STPHONECAPS;
typedef struct STPHONECAPS*		LPSTPHONECAPS;
struct STPHONECAPS
{
	DWORD	dwPhoneSwitchState;
};


// PHONESTATUS DevSpecific
typedef struct STPHONESTATUS	STPHONESTATUS;
typedef struct STPHONESTATUS*	LPSTPHONESTATUS;
struct STPHONESTATUS
{
	DWORD	dwPhoneSwitchState;
};


////////////////////////////////////////////////////////
// TAPI Device Class Extensions

#define DCSHORLINETYPE_NULL			0
#define DCSHORLINETYPE_STATION		1			// Station
#define DCSHORLINETYPE_TRUNK		2			// Trunk
#define DCSHORLINETYPE_NETWORK		3			// Network extension (see below)
#define DCSHORLINETYPE_SPECIAL		4			// Special extension (see below)

#define DCSHORSPECIALTYPE_NULL		0
#define DCSHORSPECIALTYPE_NIGHTBELL	1			// Nightbell extension
#define DCSHORSPECIALTYPE_PAGING	2			// Paging extension

#define DCSHORTRUNKTYPE_NULL		0
#define DCSHORTRUNKTYPE_A_LS		1			// Analog Loop Start trunk
#define DCSHORTRUNKTYPE_A_DID		2			// Analog Direct Inward Dial trunk	
#define DCSHORTRUNKTYPE_D_LS		3			// Digital Loop Start trunk
#define DCSHORTRUNKTYPE_D_EM		4			// Digital E&M trunk	

#define DCSHORNETWORKTYPE_NULL		0
#define DCSHORNETWORKTYPE_AA		1			// VM Auto Attendant extension
#define DCSHORNETWORKTYPE_FWD		2			// VM Forward extension (take a message)
#define DCSHORNETWORKTYPE_LOGIN		3			// VM Login extension (pickup messages)

// "shoretel/line"
typedef struct DCSHORLINE		DCSHORLINE;
typedef struct DCSHORPORT*		LPDCSHORLINE;
struct DCSHORLINE
{
	DWORD			dwLineType;					// one of DCSHORLINETYPE_xxx
	DWORD			dwSpecialType;				// one of DCSHORSPECIALTYPE_xxx
	DWORD			dwTrunkType;				// one of DCSHORTRUNKTYPE_xxx
	DWORD			dwNetworkType;				// one of DCSHORNETWORKTYPE_xxx
	DWORD			dwSwitchID;					// SwitchID from Zin switch table
	DWORD			dwPortNum;					// Port Number (if any, else 0)
	DWORD			dwExtension;				// Extension DN expressed as DWORD (if any, else 0)
	DWORD			dwSwitchNameSize;			// Number of bytes in UNICODE SwitchName string
	DWORD			dwSwitchNameOffset;			// Offset to the SwitchName string
	DWORD			dwLineDescriptionSize;		// Number of bytes in UNICODE LineDescription string
	DWORD			dwLineDescriptionOffset;	// Offset to the LineDescription string
};

// "shoretel/call"
typedef struct DCSHORCALL		DCSHORCALL;
typedef struct DCSHORCALL*		LPDCSHORCALL;
struct DCSHORCALL
{
	DWORD			dwCallID;
	DWORD			dwNumParties;
	DWORD			dwLineDeviceID[1];
};

// "shoretel/channel"
typedef struct DCSHORCHANNEL	DCSHORCHANNEL;
typedef struct DCSHORCHANNEL*	LPDCSHORCHANNEL;
struct DCSHORCHANNEL
{
	DWORD			dwChannel;					// Call ID of the Media Channel
	IN_ADDR			iaSwitch;					// IPBX IP Address
	IN_ADDR			iaClient;					// VoiceMail Server IP Address
	DWORD			dwWaveDeviceID;				// WAV Device ID for the Media Channel
	DWORD			dwCCOID;					// CCOID of the line device
};

// "shoretel/callid"
typedef struct DCSHORCALLID		DCSHORCALLID;
typedef struct DCSHORCALLID*	LPDCSHORCALLID;
struct DCSHORCALLID
{
	WCHAR			wcSipCallID[64];	// 64 wide char SIP Call ID
	DWORD			dwCallID;			// 32-bit Call ID DWORD
	GUID			guidCallID;			// 128-bit Call ID GUID
};

// additional PHONEBUTTONFUNCTION_ constants
#define PHONEBUTTONFUNCTION_UNPARK				0x80000001
#define PHONEBUTTONFUNCTION_XFERTOVM			0x80000002
#define PHONEBUTTONFUNCTION_XFERTOAA			0x80000003
#define PHONEBUTTONFUNCTION_XFERTOMAILBOX		0x80000004
#define PHONEBUTTONFUNCTION_DIALVMLOGIN			0x80000005

#define PHONEBUTTONFUNCTION_DROP1				0x80000101
#define PHONEBUTTONFUNCTION_DROP2				0x80000102
#define PHONEBUTTONFUNCTION_ALERTLAMP			0x80000103
#define PHONEBUTTONFUNCTION_NAV_SELECT			0x80000104
#define PHONEBUTTONFUNCTION_NAV_DELETE			0x80000105
#define PHONEBUTTONFUNCTION_NAV_LEFT			0x80000106
#define PHONEBUTTONFUNCTION_NAV_UP				0x80000107
#define PHONEBUTTONFUNCTION_NAV_RIGHT			0x80000108
#define PHONEBUTTONFUNCTION_NAV_DOWN			0x80000109
