//#define TESTING
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Configuration;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using AgileSoftwareInc.Tapi2Lib;

namespace AgileSoftwareInc.TapiApp
{
  public partial class MainForm : Form
  {

    #region Custom Class Members

    Properties.Settings m_settings;

    /// <summary>
    /// Tapi Objects
    /// </summary>
    private CTapi m_CTapi;

    private CLine[] m_CLines;

    private BindingList<IncomingCall> m_calls = new BindingList<IncomingCall>();

    #endregion // Custom Class Members  
    
    public MainForm()
    {
      InitializeComponent();

      m_settings = Properties.Settings.Default;

      IntPtr AppHandle = System.IntPtr.Zero;
      uint TapiVersion = Convert.ToUInt32(m_settings.TapiVersion);

      m_CTapi = new CTapi(this.Text, TapiVersion, AppHandle,
        CTapi.LineInitializeExOptions.LINEINITIALIZEEXOPTION_USEHIDDENWINDOW);
      LineCollection col = m_CTapi.GetVoiceLinesByFilter(m_settings.TSP, CTapi.LineCallPrivilege.LINECALLPRIVILEGE_MONITOR);
      m_CLines = col.ToArray();

      m_CTapi.AppNewCall += new CTapi.AppNewCallEventHandler(MyAppNewCallEventHandler);
      m_CTapi.CallStateEvent += new CTapi.CallStateEventHandler(MyCallStateEventHandler);
      m_CTapi.LineCallInfoEvent += new CTapi.LineCallInfoEventHandler(MyLineCallInfoEventHandler);

      if ((m_CLines != null) && (m_CLines.Length > 0))
      {
        foreach (CLine line in m_CLines)
          MessageBox.Show(line.GetLineInfoString());
      }
      else
        MessageBox.Show("No voice Lines found!");
    }


    private void MainForm_Load(object sender, EventArgs e)
    {
      grdCalls.DataSource = m_calls;
    }

    #region Public Methods


    public void TransferCall(IncomingCall call, string telno)
    {
      int requestID = 0;
      CTapi.LineErrReturn retval = CTapi.BlindTransfer(call.hCall, telno, out requestID);
      string ret = ((CTapi.LineErrReturn)retval).ToString();
      MessageBox.Show("Transferring call return: " + ret, this.Text);
      if (retval > 0)
      {
        // clean up
        CTapi.DeallocateCall(call.hCall);

      }
      else
      {
        //handle the error
      }
      RemoveCall(call.hCall);
    }

    public void ShutDown()
    {
      if (this.m_CTapi != null)
      {
        m_CTapi.AppNewCall -= new CTapi.AppNewCallEventHandler(MyAppNewCallEventHandler);
        m_CTapi.CallStateEvent -= new CTapi.CallStateEventHandler(MyCallStateEventHandler);
        CTapi.LineErrReturn ret = m_CTapi.ShutDown();
        if (ret != CTapi.LineErrReturn.LINEERR_OK)
        {
          // handle error!
        }
      }
    } // ShutDown


    #endregion // Public Methods

    #region Event Handling

    public void MyAppNewCallEventHandler(Object sender, CTapi.AppNewCallEventArgs e)
    {
			CCall call = e.call;
      HandleCall(call);     
#if TESTING
      call.GetCallInfo();
      MessageBox.Show(call.GetCallInfoString());
#endif
    }

    public void MyCallStateEventHandler(Object sender, CTapi.CallStateEventArgs e)
    {
      System.Diagnostics.Debug.WriteLine("Call State: " + e.call.CallID.ToString() + " Value: " + e.CallState.ToString());
#if TESTING      
      MessageBox.Show("Call State: " + CTapi.sCallState(e.CallState) + Environment.NewLine
        + "Call Handle: " + e.hcall.ToString());
#endif
      switch (e.CallState)
      {
        //case CTapi.LineCallState.LINECALLSTATE_PROCEEDING:
        //case CTapi.LineCallState.LINECALLSTATE_RINGBACK:
        case CTapi.LineCallState.LINECALLSTATE_CONNECTED:
          HandleCall(e.call);
          break;
        case CTapi.LineCallState.LINECALLSTATE_DISCONNECTED:
          RemoveCall(e.call.hCall);
          break;
        case CTapi.LineCallState.LINECALLSTATE_OFFERING:
          HandleCall(e.call);
          break;
        case CTapi.LineCallState.LINECALLSTATE_IDLE:
          RemoveCall(e.call.hCall);
          break;
      }
    }

    public void MyLineCallInfoEventHandler(object sender, CTapi.LineCallInfoEventArgs e)
    {
      CCall call = new CCall(this.m_CTapi, e.hCall);
#if TESTING
      MessageBox.Show(call.GetCallInfoString());
#endif
      HandleCall(call);
    }

    private void HandleCall(CCall call)
    {
      call.GetCallInfo();
#if TESTING
			MessageBox.Show(call.GetCallInfoString());
#endif
      string callerID = call.CallerID;
      string callerName = call.CallerIDName;
      // check if call already exists
      IncomingCall newCall = FindCall(call.hCall);
      if (newCall != null)
      {
        newCall.CallerID = callerID;
        newCall.CallerName = callerName;
        return;
      }

      newCall = new IncomingCall(call.hCall, call.CallerID, call.CallerIDName);
      AddCall(newCall);

    }


    private void AddCall(IncomingCall call)
    {
      IncomingCall c = FindCall(call.hCall);
      if (c == null)
        m_calls.Add(call);
    }

    private IncomingCall FindCall(IntPtr hCall)
    {
      foreach (IncomingCall call in m_calls)
      {
        if (call.hCall == hCall)
          return call;
      }
      return null;
    }

    private void RemoveCall(IntPtr hCall)
    {
      foreach (IncomingCall call in m_calls)
      {
        if (call.hCall == hCall)
        {
          m_calls.Remove(call);
          //grdCalls.Refresh();
          return;
        }
      }
    }

    #endregion Event Handling

    private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
    {
      m_CTapi.AppNewCall -= new CTapi.AppNewCallEventHandler(MyAppNewCallEventHandler);
      m_CTapi.CallStateEvent -= new CTapi.CallStateEventHandler(MyCallStateEventHandler);
      m_CTapi.LineCallInfoEvent -= new CTapi.LineCallInfoEventHandler(MyLineCallInfoEventHandler);
      m_CTapi.ShutDown();
    }

    private void cmdCall_Click(object sender, EventArgs e)
    {
      IntPtr hCall = IntPtr.Zero;
      int requestID = 0;
      m_CTapi.DialNumber(m_CLines[0], txtCall.Text, out hCall, out requestID);

    }

		
  }
}